"""Create pickled fig for use in integration tests."""

import pickle
from pathlib import Path

import datashader as ds
import datashader.transfer_functions as tf
import pandas as pd
import plotly.express as px
import zstandard as zstd
from colorcet import fire

cctx = zstd.ZstdCompressor(level=20)

df = pd.read_csv(
    "https://raw.githubusercontent.com/plotly/datasets/master/uber-rides-data1.csv",
)
dff = (
    df.query("Lat < 40.82")
    .query("Lat > 40.70")
    .query("Lon > -74.02")
    .query("Lon < -73.91")
)


cvs = ds.Canvas(plot_width=1000, plot_height=1000)
agg = cvs.points(dff, x="Lon", y="Lat")
# agg is an xarray object, see http://xarray.pydata.org/en/stable/ for more details
coords_lat, coords_lon = agg.coords["Lat"].to_numpy(), agg.coords["Lon"].to_numpy()
# Corners of the image
coordinates = [
    [coords_lon[0], coords_lat[0]],
    [coords_lon[-1], coords_lat[0]],
    [coords_lon[-1], coords_lat[-1]],
    [coords_lon[0], coords_lat[-1]],
]


img = tf.shade(agg, cmap=fire)[::-1].to_pil()


# Trick to create rapidly a figure with map axes
fig = px.scatter_map(dff[:1], lat="Lat", lon="Lon", zoom=12)
# Add the datashader image as a tile map layer image
fig.update_layout(
    map_style="carto-darkmatter",
    map_layers=[{"sourcetype": "image", "source": img, "coordinates": coordinates}],
)

raw = pickle.dumps(fig, protocol=5)  # >=3.8
compressed = cctx.compress(raw)
with Path(f"./figs/{Path(__file__).stem}.pkl.zst").open("wb") as f:
    f.write(compressed)

print(  # noqa: T201
    f"{Path(__file__).stem}.pkl: "
    f"{len(raw) / 1024:.1f} -> {len(compressed) / 1024:.1f} KB",
)



"""A list of errors available from the kaleido package."""

from choreographer.errors import (
    BrowserClosedError,
    BrowserFailedError,
    ChromeNotFoundError,
)

from ._kaleido_tab import JavascriptError, KaleidoError

__all__ = [
    "BrowserClosedError",
    "BrowserFailedError",
    "ChromeNotFoundError",
    "JavascriptError",
    "KaleidoError",
]



# /// script
# requires-python = ">=3.8"
# dependencies = [
#    "pandas",
#    "plotly[express]",
#    "kaleido @ file:///${PROJECT_ROOT}/",
#    ]
# ///
"""Runs the examples in the documentation. Use `mkdir output/; uv run`."""

import asyncio

import plotly.express as px

import kaleido

### SAMPLE DATA ###

fig = px.scatter(
    px.data.iris(),
    x="sepal_length",
    y="sepal_width",
    color="species",
)

fig2 = px.line(
    px.data.gapminder().query("country=='Canada'"),
    x="year",
    y="lifeExp",
    title="Life expectancy in Canada",
)

figures = [fig, fig2]

### WRITE FIGURES ###

# Simple one image synchronous write

kaleido.write_fig_sync(fig, path="./output/")


# Multiple image write with error collection

error_log = []

kaleido.write_fig_sync(
    figures,
    path="./output/",
    opts={"format": "jpg"},
    error_log=error_log,
)

# Dump the error_log

if error_log:
    for e in error_log:
        print(str(e))  # noqa: T201
    raise RuntimeError("{len(error_log)} images failed.")


# async/await style of above

asyncio.run(
    kaleido.write_fig(
        figures,
        path="./output/",
        opts={"format": "jpg"},
        error_log=error_log,
    ),
)

### Make a figure generator


def generate_figures():  # can be async as well
    """Generate plotly figures for each country in gapminder."""
    data = px.data.gapminder()
    for country in data["country"].unique():  # list all countries in dataset
        # yield unique plot for each country
        yield px.line(
            data.query(f'country=="{country}"'),
            x="year",
            y="lifeExp",
            title=f"Life expectancy in {country}",
        )


kaleido.write_fig_sync(generate_figures(), path="./output/", n=15)
# file names will be taken from figure title


### If you need more control, use an object


def generate_figure_objects():
    """Generate plotly figure objects for each country in gapminder."""
    data = px.data.gapminder()
    for country in data["country"].unique():  # list all countries in dataset
        fig = px.line(
            data.query(f'country=="{country}"'),
            x="year",
            y="lifeExp",
            title=f"Life expectancy in {country}",
        )
        yield {"fig": fig, "path": f"./output/{country}.jpg"}
        # customize file name


# use 15 processes
kaleido.write_fig_from_object_sync(generate_figure_objects(), n=15)



import pytest

from kaleido._sync_server import GlobalKaleidoServer


class TestGlobalKaleidoServer:
    """Test the GlobalKaleidoServer singleton class."""

    def test_singleton_behavior(self):
        """Test that creating the object twice returns the same instance."""
        # Should be the same object
        assert GlobalKaleidoServer() is GlobalKaleidoServer()

    def test_is_running_open_close_cycle(self):
        """Test is_running, open, and close in a loop three times."""
        server = GlobalKaleidoServer()

        # Initial state should be not running
        assert not server.is_running()

        for i in range(2):
            # Check not running
            assert not server.is_running(), (
                f"Iteration {i}: Should not be running initially"
            )

            server.open()

            # Check is running
            assert server.is_running(), f"Iteration {i}: Should be running after open"

            # Call open again - should warn
            with pytest.warns(RuntimeWarning, match="Server already open"):
                server.open()

            server.open(silence_warnings=True)
            server.close()

            # Call close again - should warn
            with pytest.warns(RuntimeWarning, match="Server already closed"):
                server.close()

            server.close(silence_warnings=True)

            # Check not running
            assert not server.is_running(), (
                f"Iteration {i}: Should not be running after close"
            )

    def test_call_function_when_not_running_raises_error(self):
        """Test that calling function when server is not running raises RuntimeError."""
        server = GlobalKaleidoServer()

        # Ensure server is closed
        if server.is_running():
            server.close(silence_warnings=True)

        # Should raise RuntimeError
        with pytest.raises(RuntimeError, match="Can't call function on stopped server"):
            server.call_function("some_function")

    def test_getattr_call_function_integration(self):
        """Test __getattr__ integration with call_function."""
        server = GlobalKaleidoServer()
        method_name = "random_method"
        test_args = ("arg1", "arg2")
        test_kwargs = {"kwarg1": "value1"}

        def method_checker(_self, name):
            assert name == method_name

            def dummy_method(*args, **kwargs):
                assert args == test_args
                assert kwargs == test_kwargs

            return dummy_method

        # Temporarily add __getattr__ to the class
        GlobalKaleidoServer.__getattr__ = method_checker

        try:
            # Call a random method with some args and kwargs
            server.random_method(*test_args, **test_kwargs)

        finally:
            # Clean up - remove __getattr__ from the class
            delattr(GlobalKaleidoServer, "__getattr__")

    def teardown_method(self):
        """Clean up after each test."""
        server = GlobalKaleidoServer()
        if server.is_running():
            server.close(silence_warnings=True)



"""
Kaleido is a library for generating static images from Plotly figures.

Please see the README.md for more information and a quickstart.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from choreographer.cli import get_chrome, get_chrome_sync

from . import _sync_server
from ._page_generator import PageGenerator
from .kaleido import Kaleido

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, Iterable
    from pathlib import Path
    from typing import Any, TypeVar, Union

    from ._fig_tools import Figurish, LayoutOpts

    T = TypeVar("T")
    AnyIterable = Union[AsyncIterable[T], Iterable[T]]

__all__ = [
    "Kaleido",
    "PageGenerator",
    "calc_fig",
    "calc_fig_sync",
    "get_chrome",
    "get_chrome_sync",
    "start_sync_server",
    "stop_sync_server",
    "write_fig",
    "write_fig_from_object",
    "write_fig_from_object_sync",
    "write_fig_sync",
]

_global_server = _sync_server.GlobalKaleidoServer()


def start_sync_server(*args: Any, silence_warnings: bool = False, **kwargs: Any):
    """
    Start a kaleido server which will process all sync generation requests.

    The kaleido server is a singleton, so it can't be opened twice. This
    function will warn you if the server is already running.

    This wrapper function takes the exact same arguments as kaleido.Kaleido(),
    except one extra, `silence_warnings`.

    Args:
        *args: all arguments `Kaleido()` would take.
        silence_warnings: (bool, default False): If True, don't emit warning if
        starting an already started server.
        **kwargs: all keyword arguments `Kaleido()` would take.

    """
    _global_server.open(*args, silence_warnings=silence_warnings, **kwargs)


def stop_sync_server(*, silence_warnings: bool = False):
    """
    Stop the kaleido server. It can be restarted. Warns if not started.

    Args:
        silence_warnings: (bool, default False): If True, don't emit warning if
        stopping a server that's not running.

    """
    _global_server.close(silence_warnings=silence_warnings)


async def calc_fig(
    fig: Figurish,
    path: str | None | Path = None,
    opts: LayoutOpts | None = None,
    *,
    topojson: str | None = None,
    kopts: dict[str, Any] | None = None,
):
    """
    Return binary for plotly figure.

    A convenience wrapper for `Kaleido.calc_fig()` which starts a `Kaleido` and
    executes the `calc_fig()`.
    It takes an additional argument, `kopts`, a dictionary of arguments to pass
    to the kaleido process. See the `kaleido.Kaleido` docs. However,
    `calc_fig()` will never use more than one processor, so any `n` value will
    be overridden.


    See documentation for `Kaleido.calc_fig()`.

    """
    kopts = kopts or {}
    kopts["n"] = 1  # should we force this?
    async with Kaleido(**kopts) as k:
        return await k.calc_fig(
            fig,
            path=path,
            opts=opts,
            topojson=topojson,
        )


async def write_fig(
    fig: Figurish,
    path: str | None | Path = None,
    opts: LayoutOpts | None = None,
    *,
    topojson: str | None = None,
    kopts: dict[str, Any] | None = None,
    **kwargs,
):
    """
    Write a plotly figure(s) to a file.

    A convenience wrapper for `Kaleido.write_fig()` which starts a `Kaleido` and
    executes the `write_fig()`.
    It takes an additional argument, `kopts`, a dictionary of arguments to pass
    to the kaleido process. See the `kaleido.Kaleido` docs.


    See documentation for `Kaleido.write_fig()` for the other arguments.

    """
    async with Kaleido(**(kopts or {})) as k:
        await k.write_fig(
            fig,
            path=path,
            opts=opts,
            topojson=topojson,
            **kwargs,
        )


async def write_fig_from_object(
    generator: AnyIterable,  # this could be more specific with []
    *,
    kopts: dict[str, Any] | None = None,
    **kwargs,
):
    """
    Write a plotly figure(s) to a file.

    A convenience wrapper for `Kaleido.write_fig_from_object()` which starts a
    `Kaleido` and executes the `write_fig_from_object()`
    It takes an additional argument, `kopts`, a dictionary of arguments to pass
    to the kaleido process. See the `kaleido.Kaleido` docs.

    See documentation for `Kaleido.write_fig_from_object()` for the other
    arguments.

    """
    async with Kaleido(**(kopts or {})) as k:
        await k.write_fig_from_object(
            generator,
            **kwargs,
        )


def calc_fig_sync(*args: Any, **kwargs: Any):
    """Call `calc_fig` but blocking."""
    if _global_server.is_running():
        return _global_server.call_function("calc_fig", *args, **kwargs)
    else:
        return _sync_server.oneshot_async_run(calc_fig, args=args, kwargs=kwargs)


def write_fig_sync(*args: Any, **kwargs: Any):
    """Call `write_fig` but blocking."""
    if _global_server.is_running():
        _global_server.call_function("write_fig", *args, **kwargs)
    else:
        _sync_server.oneshot_async_run(write_fig, args=args, kwargs=kwargs)


def write_fig_from_object_sync(*args: Any, **kwargs: Any):
    """Call `write_fig_from_object` but blocking."""
    if _global_server.is_running():
        _global_server.call_function("write_fig_from_object", *args, **kwargs)
    else:
        _sync_server.oneshot_async_run(write_fig_from_object, args=args, kwargs=kwargs)



from pathlib import Path

import pytest

from kaleido import _fig_tools

sources = ["argument", "layout", "template", "default"]
values = [None, 150, 800, 1500]


@pytest.mark.parametrize("width_source", sources)
@pytest.mark.parametrize("height_source", sources)
@pytest.mark.parametrize("width_value", values)
@pytest.mark.parametrize("height_value", [x * 1.5 if x else x for x in values])
def test_get_figure_dimensions(width_source, height_source, width_value, height_value):
    """Test _get_figure_dimensions with all combinations of width/height sources."""

    layout = {}
    width_arg = None
    expected_width = width_value

    if width_source == "argument":
        width_arg = width_value
    elif width_source == "layout":
        layout["width"] = width_value
    elif width_source == "template":
        layout.setdefault("template", {}).setdefault("layout", {})["width"] = (
            width_value
        )
    else:  # default
        expected_width = None

    # Set to default if None
    if expected_width is None:
        expected_width = _fig_tools.DEFAULT_WIDTH

    # Do for height what I did for width
    height_arg = None
    expected_height = height_value

    if height_source == "argument":
        height_arg = height_value
    elif height_source == "layout":
        layout["height"] = height_value
    elif height_source == "template":
        layout.setdefault("template", {}).setdefault("layout", {})["height"] = (
            height_value
        )
    else:  # default
        expected_height = None

    # Set to default if None
    if expected_height is None:
        expected_height = _fig_tools.DEFAULT_HEIGHT

    # Call the function
    r_width, r_height = _fig_tools._get_figure_dimensions(  # noqa: SLF001
        layout,
        width_arg,
        height_arg,
    )

    # Assert results
    assert r_width == expected_width, (
        f"Width mismatch: got {r_width}, expected {expected_width}, "
        f"source: {width_source}, value: {width_value}"
    )
    assert r_height == expected_height, (
        f"Height mismatch: got {r_height}, expected {expected_height}, "
        f"source: {height_source}, value: {height_value}"
    )


def test_next_filename_no_existing_files(tmp_path):
    """Test _next_filename when no files exist."""
    result = _fig_tools._next_filename(tmp_path, "test", "png")  # noqa: SLF001
    assert result == "test.png"


def test_next_filename_base_file_exists(tmp_path):
    """Test _next_filename when base file exists."""
    # Create the base file
    (tmp_path / "test.png").touch()

    result = _fig_tools._next_filename(tmp_path, "test", "png")  # noqa: SLF001
    assert result == "test-2.png"


def test_next_filename_numbered_files_exist(tmp_path):
    """Test _next_filename when numbered files exist."""
    # Create various numbered files
    (tmp_path / "test.png").touch()
    (tmp_path / "test-2.png").touch()
    (tmp_path / "test-3.png").touch()
    (tmp_path / "test-5.png").touch()  # Gap in numbering

    result = _fig_tools._next_filename(tmp_path, "test", "png")  # noqa: SLF001
    assert result == "test-6.png"  # Should be max + 1


def test_next_filename_similar_names_ignored(tmp_path):
    """Test _next_filename ignores files with similar but different names."""
    # Create files that shouldn't match the pattern
    (tmp_path / "test.png").touch()
    (tmp_path / "test-2.png").touch()
    (tmp_path / "testing-3.png").touch()  # Different prefix
    (tmp_path / "test-2.jpg").touch()  # Different extension
    (tmp_path / "test-abc.png").touch()  # Non-numeric suffix

    result = _fig_tools._next_filename(tmp_path, "test", "png")  # noqa: SLF001
    assert result == "test-3.png"  # Should only count test.png and test-2.png


def test_next_filename_special_characters(tmp_path):
    """Test _next_filename with special characters in prefix and extension."""
    prefix = "test-file_name"
    ext = "svg"  # set up to be parameterized but not

    # Create some files
    (tmp_path / f"{prefix}.{ext}").touch()
    (tmp_path / f"{prefix}-2.{ext}").touch()

    result = _fig_tools._next_filename(tmp_path, prefix, ext)  # noqa: SLF001
    assert result == f"{prefix}-3.{ext}"


def test_next_filename_only_numbered_files(tmp_path):
    """Test _next_filename when only numbered files exist (no base file)."""
    # Create only numbered files, no base file
    (tmp_path / "test-2.png").touch()
    (tmp_path / "test-3.png").touch()
    (tmp_path / "test-10.png").touch()

    result = _fig_tools._next_filename(tmp_path, "test", "png")  # noqa: SLF001
    assert result == "test-11.png"  # Should be max + 1


# Fixtures for _build_full_path tests - testing various title scenarios
@pytest.fixture(
    params=[
        (
            {
                "layout": {
                    "title": {"text": "My-Test!@#$%^&*()Chart_with[lots]of{symbols}"},
                },
            },
            "My_TestChart_withlotsofsymbols",
        ),  # Complex title
        (
            {"layout": {"title": {"text": "Simple Title"}}},
            "Simple_Title",
        ),  # Simple title
        ({"layout": {}}, "fig"),  # No title
    ],
)
def fig_fixture(request):
    """Parameterized fixture for fig with various title scenarios."""
    return request.param


def test_build_full_path_no_path_input(fig_fixture):
    """Test _build_full_path with no path input uses current path."""
    fig_dict, expected_prefix = fig_fixture
    result = _fig_tools._build_full_path(None, fig_dict, "ext")  # noqa: SLF001

    # Should use current directory
    assert result.parent.resolve() == Path().cwd().resolve()
    assert result.parent.is_dir()

    assert result.name == f"{expected_prefix}.ext"


def test_build_full_path_no_suffix_directory(tmp_path, fig_fixture):
    """Test _build_full_path with path having no suffix."""
    fig_dict, expected_prefix = fig_fixture

    # Test directory no suffix
    test_dir = tmp_path
    result = _fig_tools._build_full_path(test_dir, fig_dict, "ext")  # noqa: SLF001

    # Should use provided directory
    assert result.parent == test_dir
    assert result.name == f"{expected_prefix}.ext"

    # Test error
    nonexistent_dir = Path("/nonexistent/directory")
    with pytest.raises(ValueError, match=r"Directory .* not found. Please create it."):
        _fig_tools._build_full_path(nonexistent_dir, fig_dict, "ext")  # noqa: SLF001


def test_build_full_path_directory_with_suffix(tmp_path, fig_fixture):
    """Test _build_full_path with path that is directory even with suffix."""
    fig_dict, expected_prefix = fig_fixture

    # Create a directory with a suffix-like name
    dir_with_suffix = tmp_path / "mydir.png"
    dir_with_suffix.mkdir()

    result = _fig_tools._build_full_path(dir_with_suffix, fig_dict, "ext")  # noqa: SLF001

    # Should treat as directory
    assert result.parent == dir_with_suffix
    assert result.name == f"{expected_prefix}.ext"


def test_build_full_path_file_with_suffix(tmp_path, fig_fixture):
    """Test _build_full_path with file path having suffix."""
    fig_dict, _expected_prefix = fig_fixture

    # Exists
    file_path = tmp_path / "output.png"
    result = _fig_tools._build_full_path(file_path, fig_dict, "ext")  # noqa: SLF001

    # Should return the exact path provided
    assert result == file_path

    # Doesn't exist
    file_path = Path("/nonexistent/directory/output.png")
    with pytest.raises(
        RuntimeError,
        match=r"Cannot reach path .* Are all directories created?",
    ):
        _fig_tools._build_full_path(file_path, fig_dict, "ext")  # noqa: SLF001



import logging
import os
import platform

import logistro
import pytest
from hypothesis import HealthCheck, settings

_logger = logistro.getLogger(__name__)

# we turn off deadlines for macs and windows in CI
# i'm under the impression that they are resource limited
# and delays are often caused by sharing resources with other
# virtualization neighbors.

settings.register_profile(
    "ci",
    deadline=None,  # no per-example deadline
    suppress_health_check=(HealthCheck.too_slow,),  # avoid flaky "too slow" on CI
)

is_ci = os.getenv("GITHUB_ACTIONS") == "true" or os.getenv("CI") == "true"
if is_ci and platform.system() in {"Windows", "Darwin"}:
    settings.load_profile("ci")


# pytest shuts down its capture before logging/threads finish
@pytest.fixture(scope="session", autouse=True)
def cleanup_logging_handlers(request):
    capture = request.config.getoption("--capture") != "no"
    try:
        yield
    finally:
        if capture:
            _logger.info("Conftest cleaning up handlers.")
            for handler in logging.root.handlers[:]:
                handler.flush()
                if isinstance(handler, logging.StreamHandler):
                    logging.root.removeHandler(handler)



import numpy as np
import plotly.graph_objects as go
import pytest
from plotly.subplots import make_subplots

import kaleido

pytestmark = pytest.mark.asyncio(loop_scope="function")

rng = np.random.default_rng()  # creates a Generator instance


async def test_complex_plotly_encoder():
    """Test that kaleido can handle complex Plotly figures with numpy arrays."""

    # Create complex numpy arrays
    x = np.linspace(0, 4 * np.pi, 100)
    y1 = np.sin(x) * np.exp(-x / 10)

    # Create a 2D array for heatmap
    z = np.outer(
        np.sin(np.linspace(0, np.pi, 20)),
        np.cos(np.linspace(0, np.pi, 30)),
    )

    # Create subplot figure
    fig = make_subplots(
        rows=2,
        cols=2,
        subplot_titles=(
            "Complex Scatter",
            "Heatmap",
            "Bar with Color",
            "3D Surface",
        ),
        specs=[
            [{"type": "scatter"}, {"type": "heatmap"}],
            [{"type": "bar"}, {"type": "surface"}],
        ],
    )

    # Scatter with numpy marker sizes and complex styling
    fig.add_trace(
        go.Scatter(
            x=x,
            y=y1,
            mode="lines+markers",
            line={"color": "blue", "width": 2},
            marker={
                "size": [rng.integers(2, 12) for _ in range(len(x))],
                "color": np.abs(y1),
                "colorscale": "Viridis",
                "showscale": True,
            },
        ),
        row=1,
        col=1,
    )

    # Heatmap with 2D numpy array
    fig.add_trace(
        go.Heatmap(
            z=z,
            colorscale="Plasma",
        ),
        row=1,
        col=2,
    )

    # Bar chart with numpy data and color mapping
    categories = np.array(["A", "B", "C", "D", "E", "F"])
    values = rng.normal(50, 20, len(categories))

    fig.add_trace(
        go.Bar(
            x=categories,
            y=values,
            marker={
                "color": np.abs(values),
                "colorscale": "Blues",
                "line": {"width": 2, "color": "black"},
            },
        ),
        row=2,
        col=1,
    )

    # 3D surface with complex numpy operations
    x_surf = np.linspace(-3, 3, 30)
    y_surf = np.linspace(-3, 3, 30)
    X, Y = np.meshgrid(x_surf, y_surf)  # noqa: N806
    Z = np.sin(np.sqrt(X**2 + Y**2)) * np.exp(-(X**2 + Y**2) / 10)  # noqa: N806

    fig.add_trace(
        go.Surface(
            x=X,
            y=Y,
            z=Z,
            colorscale="Cividis",
        ),
        row=2,
        col=2,
    )

    # Complex layout with numpy-based annotations
    fig.update_layout(
        title="Complex Numpy Figure for Encoder Testing",
        height=800,
        width=1200,
    )

    # Render with kaleido
    img_bytes = await kaleido.calc_fig(fig)

    assert isinstance(img_bytes, bytes)



"""Tests for wrapper functions in __init__.py that test argument passing."""

from unittest.mock import AsyncMock, patch

import pytest

import kaleido

# Pretty complicated for basically testing a bunch of wrappers, but it works.
# Integration tests seem more important.
# I much prefer the public_api file, this set of tests can be considered
# for deletion.


@pytest.fixture
def args():
    """Basic args for sync wrapper tests."""
    return ({"data": []}, "test.png")


@pytest.fixture
def kwargs():
    """Basic kwargs for sync wrapper tests."""
    return {"width": 800}


@patch("kaleido._sync_server.GlobalKaleidoServer.open")
def test_start_sync_server_passes_args(mock_open, args, kwargs):
    """Test that start_sync_server passes args and silence_warnings correctly."""
    # Test with silence_warnings=False (default)
    kaleido.start_sync_server(*args, **kwargs)
    mock_open.assert_called_with(*args, silence_warnings=False, **kwargs)

    # Reset mock and test with silence_warnings=True
    mock_open.reset_mock()
    kaleido.start_sync_server(*args, silence_warnings=True, **kwargs)
    mock_open.assert_called_with(*args, silence_warnings=True, **kwargs)


@patch("kaleido._sync_server.GlobalKaleidoServer.close")
def test_stop_sync_server_passes_args(mock_close):
    """Test that stop_sync_server passes silence_warnings correctly."""
    # Test with silence_warnings=False (default)
    kaleido.stop_sync_server()
    mock_close.assert_called_with(silence_warnings=False)

    # Reset mock and test with silence_warnings=True
    mock_close.reset_mock()
    kaleido.stop_sync_server(silence_warnings=True)
    mock_close.assert_called_with(silence_warnings=True)


@patch("kaleido.Kaleido")
async def test_async_wrapper_functions(mock_kaleido_class):
    """Test all async wrapper functions pass arguments correctly.

    Note: This test uses fixed args rather than fixtures due to specific
    requirements with topojson and kopts that don't match the simple fixture pattern.
    """
    # Create a mock that doesn't need the context fixture
    mock_kaleido_class.return_value = mock_kaleido = AsyncMock()
    mock_kaleido.__aenter__.return_value = mock_kaleido
    mock_kaleido.__aexit__.return_value = None
    mock_kaleido.calc_fig.return_value = b"test_bytes"

    fig = {"data": []}

    # Test calc_fig with full arguments and kopts forcing n=1
    path = "test.png"
    opts = {"width": 800}
    topojson = "test_topojson"
    kopts = {"some_option": "value"}

    result = await kaleido.calc_fig(fig, path, opts, topojson=topojson, kopts=kopts)

    expected_kopts = {"some_option": "value", "n": 1}
    mock_kaleido_class.assert_called_with(**expected_kopts)
    mock_kaleido.calc_fig.assert_called_with(
        fig,
        path=path,
        opts=opts,
        topojson=topojson,
    )
    assert result == b"test_bytes"

    # Reset mocks
    mock_kaleido_class.reset_mock()
    mock_kaleido.calc_fig.reset_mock()

    # Test calc_fig with empty kopts
    await kaleido.calc_fig(fig)
    mock_kaleido_class.assert_called_with(n=1)

    # Reset mocks
    mock_kaleido_class.reset_mock()
    mock_kaleido.write_fig.reset_mock()

    # Test write_fig with full arguments
    await kaleido.write_fig(fig, path, opts, topojson=topojson, kopts=kopts)
    mock_kaleido_class.assert_called_with(**kopts)  # write_fig doesn't force n=1
    mock_kaleido.write_fig.assert_called_with(
        fig,
        path=path,
        opts=opts,
        topojson=topojson,
    )

    # Reset mocks
    mock_kaleido_class.reset_mock()
    mock_kaleido.write_fig.reset_mock()

    # Test write_fig with empty kopts
    await kaleido.write_fig(fig)
    mock_kaleido_class.assert_called_with()

    # Reset mocks
    mock_kaleido_class.reset_mock()
    mock_kaleido.write_fig_from_object.reset_mock()

    # Test write_fig_from_object
    generator = [{"data": []}]
    await kaleido.write_fig_from_object(generator, kopts=kopts)
    mock_kaleido_class.assert_called_with(**kopts)
    mock_kaleido.write_fig_from_object.assert_called_with(generator)


@patch("kaleido._sync_server.GlobalKaleidoServer.is_running")
@patch("kaleido._sync_server.GlobalKaleidoServer.call_function")
def test_sync_wrapper_server(mock_call_function, mock_is_running, args, kwargs):
    """Test all sync wrapper functions when global server is running."""
    mock_is_running.return_value = True

    # Test calc_fig_sync
    kaleido.calc_fig_sync(*args, **kwargs)
    mock_call_function.assert_called_with("calc_fig", *args, **kwargs)

    mock_call_function.reset_mock()

    # Test write_fig_sync
    kaleido.write_fig_sync(*args, **kwargs)
    mock_call_function.assert_called_with("write_fig", *args, **kwargs)

    mock_call_function.reset_mock()

    # Test write_fig_from_object_sync
    kaleido.write_fig_from_object_sync(*args, **kwargs)
    mock_call_function.assert_called_with("write_fig_from_object", *args, **kwargs)


@patch("kaleido._sync_server.GlobalKaleidoServer.is_running")
@patch("kaleido._sync_server.oneshot_async_run")
def test_sync_wrapper_oneshot(mock_oneshot_run, mock_is_running, args, kwargs):
    """Test all sync wrapper functions when no server is running."""
    mock_is_running.return_value = False

    # Test calc_fig_sync
    kaleido.calc_fig_sync(*args, **kwargs)
    mock_oneshot_run.assert_called_with(kaleido.calc_fig, args=args, kwargs=kwargs)

    mock_oneshot_run.reset_mock()

    # Test write_fig_sync
    kaleido.write_fig_sync(*args, **kwargs)
    mock_oneshot_run.assert_called_with(kaleido.write_fig, args=args, kwargs=kwargs)

    mock_oneshot_run.reset_mock()

    # Test write_fig_from_object_sync
    kaleido.write_fig_from_object_sync(*args, **kwargs)
    mock_oneshot_run.assert_called_with(
        kaleido.write_fig_from_object,
        args=args,
        kwargs=kwargs,
    )



"""Integrative tests for all public API functions in __init__.py using basic figures."""

from pathlib import Path
from unittest.mock import patch

import pytest

import kaleido


@pytest.fixture(params=["figure", "dict"])
def simple_figure(request):
    """Create a simple plotly figure for testing, either as figure or dict."""
    # ruff: noqa: PLC0415
    import plotly.express as px

    fig = px.line(x=[1, 2, 3, 4], y=[1, 2, 3, 4])

    if request.param == "dict":
        return fig.to_dict()
    return fig


async def test_async_api_functions(simple_figure, tmp_path):
    """Test calc_fig, write_fig, and write_fig_from_object with cross-validation."""
    # Test calc_fig and get reference bytes
    calc_result = await kaleido.calc_fig(simple_figure)
    assert isinstance(calc_result, bytes)
    assert calc_result.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"

    # Test write_fig and compare with calc_fig output
    write_fig_output = tmp_path / "test_write_fig.png"
    await kaleido.write_fig(simple_figure, path=str(write_fig_output))

    with Path(write_fig_output).open("rb") as f:  # noqa: ASYNC230
        write_fig_bytes = f.read()

    assert write_fig_bytes == calc_result
    assert write_fig_bytes.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"

    # Test write_fig_from_object and compare with calc_fig output
    write_fig_from_object_output = tmp_path / "test_write_fig_from_object.png"
    await kaleido.write_fig_from_object(
        [
            {
                "fig": simple_figure,
                "path": write_fig_from_object_output,
            },
        ],
    )

    with Path(write_fig_from_object_output).open("rb") as f:  # noqa: ASYNC230
        write_fig_from_object_bytes = f.read()

    assert write_fig_from_object_bytes == calc_result
    assert write_fig_from_object_bytes.startswith(
        b"\x89PNG\r\n\x1a\n",
    ), "Not a PNG file"

    # Cross-validate all results are identical
    assert write_fig_bytes == write_fig_from_object_bytes == calc_result


async def test_sync_api_functions(simple_figure, tmp_path):
    """Test sync wrappers with cross-validation."""
    # Get expected bytes from calc_fig for comparison
    expected_bytes = await kaleido.calc_fig(simple_figure)
    assert isinstance(expected_bytes, bytes)
    assert expected_bytes.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"

    # Test scenario 1: server running
    write_fig_output_1 = tmp_path / "test_write_fig_server_running.png"
    write_fig_from_object_output_1 = tmp_path / "test_from_object_server_running.png"

    with patch(
        "kaleido._sync_server.oneshot_async_run",
        wraps=kaleido._sync_server.oneshot_async_run,  # noqa: SLF001 internal
    ) as mock_oneshot, patch(
        "kaleido._global_server.call_function",
        wraps=kaleido._global_server.call_function,  # noqa: SLF001 internal
    ) as mock_call:
        kaleido.start_sync_server(silence_warnings=True)
        try:
            # Test calc_fig_sync
            calc_result_1 = kaleido.calc_fig_sync(simple_figure)
            assert isinstance(calc_result_1, bytes)
            assert calc_result_1.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"
            assert calc_result_1 == expected_bytes

            # Test write_fig_sync
            kaleido.write_fig_sync(simple_figure, path=str(write_fig_output_1))

            with Path(write_fig_output_1).open("rb") as f:  # noqa: ASYNC230
                write_fig_bytes_1 = f.read()
            assert write_fig_bytes_1 == expected_bytes
            assert write_fig_bytes_1.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"

            # Test write_fig_from_object_sync
            kaleido.write_fig_from_object_sync(
                [
                    {
                        "fig": simple_figure,
                        "path": write_fig_from_object_output_1,
                    },
                ],
            )

            with Path(write_fig_from_object_output_1).open("rb") as f:  # noqa: ASYNC230
                from_object_bytes_1 = f.read()
            assert from_object_bytes_1 == expected_bytes
            assert from_object_bytes_1.startswith(
                b"\x89PNG\r\n\x1a\n",
            ), "Not a PNG file"

            # Should have been called three times (once for each function)
            assert mock_call.call_count == 3  # noqa: PLR2004
            assert mock_oneshot.call_count == 0

            # Cross-validate all server running results are identical
            assert (
                calc_result_1
                == write_fig_bytes_1
                == from_object_bytes_1
                == expected_bytes
            )

        finally:
            kaleido.stop_sync_server(silence_warnings=True)

        # Test scenario 2: server not running
        write_fig_output_2 = tmp_path / "test_write_fig_server_not_running.png"
        write_fig_from_object_output_2 = (
            tmp_path / "test_from_object_server_not_running.png"
        )

        # Test calc_fig_sync
        calc_result_2 = kaleido.calc_fig_sync(simple_figure)
        assert isinstance(calc_result_2, bytes)
        assert calc_result_2.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"
        assert calc_result_2 == expected_bytes

        # Test write_fig_sync
        kaleido.write_fig_sync(simple_figure, path=str(write_fig_output_2))

        with Path(write_fig_output_2).open("rb") as f:  # noqa: ASYNC230
            write_fig_bytes_2 = f.read()
        assert write_fig_bytes_2 == expected_bytes
        assert write_fig_bytes_2.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"

        # Test write_fig_from_object_sync
        kaleido.write_fig_from_object_sync(
            [
                {
                    "fig": simple_figure,
                    "path": write_fig_from_object_output_2,
                },
            ],
        )

        with Path(write_fig_from_object_output_2).open("rb") as f:  # noqa: ASYNC230
            from_object_bytes_2 = f.read()
        assert from_object_bytes_2 == expected_bytes
        assert from_object_bytes_2.startswith(b"\x89PNG\r\n\x1a\n"), "Not a PNG file"

        # Should have been called three times (once for each function)
        assert mock_call.call_count == 3  # noqa: PLR2004
        assert mock_oneshot.call_count == 3  # noqa: PLR2004

        # Cross-validate all server not running results are identical
        assert (
            calc_result_2 == write_fig_bytes_2 == from_object_bytes_2 == expected_bytes
        )


def test_start_stop_sync_server_integration():
    """Test start_sync_server and stop_sync_server with warning behavior."""
    # Test starting and stopping with warnings silenced
    kaleido.start_sync_server(silence_warnings=False)

    # Test starting already started server - should warn
    with pytest.warns(RuntimeWarning, match="already"):
        kaleido.start_sync_server(silence_warnings=False)

    kaleido.start_sync_server(silence_warnings=True)

    kaleido.stop_sync_server(silence_warnings=False)

    # Test stopping already stopped server - should warn
    with pytest.warns(RuntimeWarning, match="closed"):
        kaleido.stop_sync_server(silence_warnings=False)

    kaleido.stop_sync_server(silence_warnings=True)



from __future__ import annotations

import argparse
import asyncio
import multiprocessing
import sys
import time
import warnings
from pathlib import Path
from pprint import pp
from random import sample
from typing import TypedDict

import logistro
import orjson

import kaleido

_logger = logistro.getLogger(__name__)

cpus = multiprocessing.cpu_count()

# Extract jsons of mocks
test_dir = Path(__file__).resolve().parent.parent / "integration_tests"
in_dir = test_dir / "mocks"
out_dir = test_dir / "renders"


def _get_jsons_in_paths(path: str | Path) -> list[Path]:
    # Work with Paths and directories
    path = Path(path) if isinstance(path, str) else path

    if path.is_dir():
        _logger.info(f"Input is path {path}")
        return list(path.glob("*.json"))
    elif path.is_file():
        _logger.info(f"Input is file {path}")
        return [path]
    else:
        raise TypeError("--input must be file or directory")


class Param(TypedDict):
    name: str
    opts: dict[str, int | float]


def _load_figures_from_paths(paths: list[Path]):
    # Set json
    params: list[Param]
    for path in paths:
        if path.is_file():
            with path.open(encoding="utf-8") as file:
                figure = orjson.loads(file.read())
                _logger.info(f"Yielding {path.stem}")
                if args.parameterize_opts is False:
                    params = [
                        {
                            "name": f"{path.stem}.{args.format or 'png'}",
                            "opts": {
                                "scale": args.scale,
                                "width": args.width,
                                "height": args.height,
                            },
                        },
                    ]
                else:
                    widths = [args.width] if args.width else [200, 700, 1000]
                    heights = [args.height] if args.height else [200, 500, 1000]
                    scales = [args.scale] if args.scale else [0.5, 1, 2]
                    formats = (
                        [args.format]
                        if args.format
                        else [
                            "png",
                            "pdf",
                            "jpg",
                            "webp",
                            "svg",
                            "json",
                        ]
                    )
                    params = []
                    for w in widths:
                        for h in heights:
                            for s in scales:
                                for f in formats:
                                    params.append(
                                        {
                                            "name": (
                                                f"{path.stem!s}-{w!s}"
                                                f"x{h!s}X{s!s}.{f!s}"
                                            ),
                                            "opts": {
                                                "scale": s,
                                                "width": w,
                                                "height": h,
                                            },
                                        },
                                    )
                for p in params:
                    yield {
                        "fig": figure,
                        "path": str(Path(args.output) / p["name"]),
                        "opts": p["opts"],
                    }
        else:
            raise RuntimeError(f"Path {path} is not a file.")


# Set the arguments
description = """kaleido_mocker will load up json files of plotly figs and export them.

If you set multiple process, -n, non-headless mode won't function well because
chrome will actually throttle tabs or windows/visibile- unless that tab/window
is headless.

The export of the program is a json object containing information about the execution.
"""

if "--headless" in sys.argv and "--no-headless" in sys.argv:
    raise ValueError(
        "Choose either '--headless' or '--no-headless'.",
    )

parser = argparse.ArgumentParser(
    add_help=True,
    parents=[logistro.parser],
    conflict_handler="resolve",
    description=description,
)
parser.add_argument(
    "--logistro-level",
    default="INFO",
    dest="log",
    help="Set the logging level (default INFO)",
)
parser.add_argument(
    "--n",
    type=int,
    default=cpus,
    help="Number of tabs, defaults to # of cpus",
)
parser.add_argument(
    "--input",
    type=str,
    default=in_dir,
    help="Directory of mock file/s or single file (default tests/mocks)",
)
parser.add_argument(
    "--output",
    type=str,
    default=out_dir,
    help="DIRECTORY of mock file/s (default tests/renders)",
)
parser.add_argument(
    "--format",
    type=str,
    default=None,
    help="png (default), pdf, jpg, webp, svg, json",
)
parser.add_argument(
    "--width",
    type=str,
    default=None,
    help="width in pixels (default 700)",
)
parser.add_argument(
    "--height",
    type=str,
    default=None,
    help="height in pixels (default 500)",
)
parser.add_argument(
    "--scale",
    type=str,
    default=None,
    help="Scale ratio, acts as multiplier for height/width (default 1)",
)
parser.add_argument(
    "--parameterize_opts",
    action="store_true",
    default=False,
    help="Run mocks w/ different configurations.",
)
parser.add_argument(
    "--timeout",
    type=int,
    default=90,
    help="Set timeout in seconds for any 1 mock (default 60 seconds)",
)
parser.add_argument(
    "--headless",
    action="store_true",
    default=True,
    help="Set headless as True (default)",
)
parser.add_argument(
    "--no-headless",
    action="store_false",
    dest="headless",
    help="Set headless as False",
)
parser.add_argument(
    "--stepper",
    action="store_true",
    default=False,
    dest="stepper",
    help="Stepper sets n to 1, headless to False, no timeout "
    "and asks for confirmation before printing.",
)
parser.add_argument(
    "--random",
    type=int,
    default=0,
    help="Will select N random jsons- or if 0 (default), all.",
)
parser.add_argument(
    "--fail-fast",
    action="store_true",
    default=False,
    help="Throw first error encountered and stop execution.",
)

args = parser.parse_args()
logistro.getLogger().setLevel(args.log)

if not Path(args.output).is_dir():
    raise ValueError(f"Specified output must be existing directory. Is {args.output!s}")


# Function to process the images
async def _main(error_log=None, profiler=None):
    paths = _get_jsons_in_paths(args.input)
    if args.random:
        if args.random > len(paths):
            raise ValueError(
                f"Input discover {len(paths)} paths, but a sampling of"
                f"{args.random} was asked for.",
            )
        paths = sample(paths, args.random)
    if args.stepper:
        _logger.info("Setting stepper.")
        args.n = 1
        args.headless = False
        args.timeout = 0
        if args.format == "svg":
            warnings.warn(
                "Stepper won't render svgs. It's feasible, "
                "but the adaption is just slightly more involved.",
                stacklevel=1,
            )
            await asyncio.sleep(3)
        # sets a global in kaleido, gross huh

    async with kaleido.Kaleido(
        page_generator=kaleido.PageGenerator(force_cdn=True),
        n=args.n,
        headless=args.headless,
        timeout=args.timeout,
        stepper=args.stepper,
    ) as k:
        await k.write_fig_from_object(
            _load_figures_from_paths(paths),
            error_log=error_log,
            profiler=profiler,
        )


def build_mocks():
    start = time.perf_counter()
    try:
        error_log = [] if not args.fail_fast else None
        profiler = {}
        asyncio.run(_main(error_log, profiler))
    finally:
        # ruff: noqa: PLC0415
        from operator import itemgetter

        for tab, tab_profile in profiler.items():
            profiler[tab] = sorted(
                tab_profile,
                key=itemgetter("duration"),
                reverse=True,
            )

        elapsed = time.perf_counter() - start
        with_error_log = error_log is not None
        results = {
            "error_log": [str(log) for log in error_log] if with_error_log else None,
            "profiles": profiler,
            "total_time": f"Time taken: {elapsed:.6f} seconds",
            "total_errors": len(error_log) if with_error_log else "untracked",
        }
        pp(results)
        if error_log:
            sys.exit(1)


if __name__ == "__main__":
    build_mocks()



from __future__ import annotations

from pathlib import Path
from urllib.parse import urlparse
from urllib.request import url2pathname

import logistro

_logger = logistro.getLogger(__name__)

DEFAULT_PLOTLY = "https://cdn.plot.ly/plotly-2.35.2.js"
DEFAULT_MATHJAX = (
    "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js"
    "?config=TeX-AMS-MML_SVG"
)

KJS_PATH = Path(__file__).resolve().parent / "vendor" / "kaleido_scopes.js"


def _ensure_path(path: Path | str | tuple[str | Path, str]) -> None:
    if isinstance(path, tuple):
        path = path[0]
    _logger.debug(f"Ensuring path {path!s}")
    parsed = urlparse(str(path))
    _logger.debug(f"Parsed file path: {parsed}")
    if parsed.scheme.startswith("http"):  # is url
        return
    elif parsed.scheme.startswith("file"):
        if (_p := Path(url2pathname(parsed.path))).exists():
            return
        _logger.error(f"File parsed to: {_p}")
    elif Path(path).exists():
        return
    raise FileNotFoundError(f"{path!s} does not exist.")


class PageGenerator:
    """
    A page generator can set the versions of the js libraries used to render.

    It does this by outputting the HTML used to render the plotly figures.
    """

    header = """
<!DOCTYPE html>
<html>
    <head>
        <style id="head-style"></style>
        <title>Kaleido-fier</title>
        <script>
          window.PlotlyConfig = {MathJaxConfig: 'local'}
        </script>
        <script type="text/x-mathjax-config">
          MathJax.Hub.Config({ "SVG": { blacker: 0 }})
        </script>
"""
    """The header is the HTML that always goes at the top. Rarely needs changing."""

    footer = f"""
        <script src="{KJS_PATH.as_uri()}"></script>
    </head>
    <body style="{{margin: 0; padding: 0;}}"><img id="kaleido-image" /></body>
</html>
"""
    """The footer is the HTML that always goes on the bottom. Rarely needs changing."""

    def __init__(  # noqa: C901
        self,
        *,
        plotly: None | Path | str | tuple[Path | str, str] = None,
        mathjax: None | Path | str | bool | tuple[Path | str, str] = None,
        others: None | list[Path | str | tuple[Path | str, str]] = None,
        force_cdn: bool = False,
    ):
        """
        Create a PageGenerator.

        Args:
            plotly: The url to the plotly.js to use. Defaults to plotly.js
                present in plotly.py, if installed. Otherwise fallback to
                value of DEFAULT_PLOTLY.
            mathjax: The url to the mathjax script. Defaults to values of
                DEFAULT_MATHJAX. Can be set to false to disable mathjax.
            others: A list of other script urls to include. Usually strings, but
                can be (str, str) where it's (url, encoding).
            force_cdn: Set True to force CDN use, defaults to False.

        """
        self._scripts = []
        if mathjax is not False:
            if mathjax is None or mathjax is True:
                mathjax = DEFAULT_MATHJAX
            elif mathjax:
                _ensure_path(mathjax)
            self._scripts.append(mathjax)
        if force_cdn:
            plotly = (DEFAULT_PLOTLY, "utf-8")
        elif not plotly:
            try:
                # ruff: noqa: PLC0415
                # is this the best way to do this? can't we use importlib?
                import plotly as pltly  # type: ignore[import-untyped]

                plotly_path = (
                    Path(pltly.__file__).parent / "package_data" / "plotly.min.js"
                )
                plotly = (
                    plotly_path.as_uri(),
                    "utf-8",
                )
                if not plotly_path.is_file():
                    _logger.warning(
                        f"Found plotly but path to js is wrong? {plotly[0]}",
                    )
                    plotly = (DEFAULT_PLOTLY, "utf-8")
                else:
                    plotly[0]
            except ImportError:
                _logger.info("Plotly not installed. Using CDN.")
                plotly = (DEFAULT_PLOTLY, "utf-8")
        elif isinstance(plotly, (str, Path)):
            _ensure_path(plotly)
            plotly = (plotly, "utf-8")
        _logger.debug(f"Plotly script: {plotly}")
        self._scripts.append(plotly)
        if others:
            for o in others:
                _ensure_path(o)
            self._scripts.extend(others)

    def generate_index(self, path=None):
        """
        Generate the page.

        Args:
            path: If specified, page is written to path. Otherwise it is returned.

        """
        page = self.header
        script_tag = '\n        <script src="%s"></script>'
        script_tag_charset = '\n        <script src="%s" charset="%s"></script>'
        for script in self._scripts:
            if isinstance(script, (str, Path)):
                page += script_tag % str(script)
            else:
                page += script_tag_charset % script
        page += self.footer
        _logger.debug2(page)
        if not path:
            return page
        with (path).open("w") as f:
            f.write(page)
        return path.as_uri()



"""
Adapted from old code, it 1. validates, 2. write defaults, 3. packages object.

Its a bit complicated and mixed in order.
"""

from __future__ import annotations

import glob
import re
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypedDict

import logistro

if TYPE_CHECKING:
    from typing import Any

    from typing_extensions import TypeGuard

    Figurish = Any  # Be nice to make it more specific, dictionary or something
    FormatString = Literal["png", "jpg", "jpeg", "webp", "svg", "json", "pdf"]

_logger = logistro.getLogger(__name__)

# constants
DEFAULT_EXT = "png"
DEFAULT_SCALE = 1
DEFAULT_WIDTH = 700
DEFAULT_HEIGHT = 500
SUPPORTED_FORMATS: tuple[FormatString, ...] = (
    "png",
    "jpg",
    "jpeg",
    "webp",
    "svg",
    "json",
    "pdf",
)


def _assert_format(ext: str) -> TypeGuard[FormatString]:
    if ext not in SUPPORTED_FORMATS:
        raise ValueError(
            f"Invalid format '{ext}'.\n    Supported formats: {SUPPORTED_FORMATS!s}",
        )
    return True


def _is_figurish(o: Any) -> TypeGuard[Figurish]:
    valid = hasattr(o, "to_dict") or (isinstance(o, dict) and "data" in o)
    if not valid:
        _logger.debug(
            f"Figure has to_dict? {hasattr(o, 'to_dict')} "
            f"is dict? {isinstance(o, dict)} "
            f"Keys: {o.keys() if hasattr(o, 'keys') else None!s}",
        )
    return valid


def _get_figure_dimensions(
    layout: dict,
    width: float | None,
    height: float | None,
) -> tuple[float, float]:
    # Compute image width / height with fallbacks
    width = (
        width
        or layout.get("width")
        or layout.get("template", {}).get("layout", {}).get("width")
        or DEFAULT_WIDTH
    )
    height = (
        height
        or layout.get("height")
        or layout.get("template", {}).get("layout", {}).get("height")
        or DEFAULT_HEIGHT
    )
    return width, height


def _get_format(extension: str) -> FormatString:
    formatted_extension = extension.lower()
    if formatted_extension == "jpg":
        return "jpeg"
    if not _assert_format(formatted_extension):
        raise ValueError  # this line will never be reached its for typer
    return formatted_extension


# Input of to_spec (user gives us this)
class LayoutOpts(TypedDict, total=False):
    format: FormatString | None
    scale: int | float
    height: int | float
    width: int | float


# Output of to_spec (we give kaleido_scopes.js this)
# refactor note: this could easily be right before send
class Spec(TypedDict):
    format: FormatString
    width: int | float
    height: int | float
    scale: int | float
    data: Figurish


# validate configuration options for kaleido.js and package like its wants
def to_spec(figure: Figurish, layout_opts: LayoutOpts) -> Spec:
    # Get figure layout
    layout = figure.get("layout", {})

    for k, v in layout_opts.items():
        if k == "format":
            if v is not None and not isinstance(v, (str)):
                raise TypeError(
                    f"{k} must be one of {SUPPORTED_FORMATS!s} or None, not {v}.",
                )
        elif k in ("scale", "height", "width"):
            if v is not None and not isinstance(v, (float, int)):
                raise TypeError(f"{k} must be numeric or None, not {v}.")
        else:
            raise AttributeError(f"Unknown key in layout options, {k}")

    # Extract info
    extension = _get_format(layout_opts.get("format") or DEFAULT_EXT)

    width, height = _get_figure_dimensions(
        layout,
        layout_opts.get("width"),
        layout_opts.get("height"),
    )
    scale = layout_opts.get("scale", DEFAULT_SCALE)

    return {
        "format": extension,
        "width": width,
        "height": height,
        "scale": scale,
        "data": figure,
    }


# if we need to suffix the filename automatically:
def _next_filename(path: Path | str, prefix: str, ext: str) -> str:
    path = path if isinstance(path, Path) else Path(path)
    default = 1 if (path / f"{prefix}.{ext}").exists() else 0
    re_number = re.compile(
        r"^" + re.escape(prefix) + r"\-(\d+)\." + re.escape(ext) + r"$",
    )
    escaped_prefix = glob.escape(prefix)
    escaped_ext = glob.escape(ext)
    numbers = [
        int(match.group(1))
        for name in path.glob(f"{escaped_prefix}-*.{escaped_ext}")
        if (match := re_number.match(Path(name).name))
    ]
    n = max(numbers, default=default) + 1
    return f"{prefix}.{ext}" if n == 1 else f"{prefix}-{n}.{ext}"


# validate and build full route if needed:
def _build_full_path(
    path: Path | None,
    fig: Figurish,
    ext: FormatString,
) -> Path:
    full_path: Path | None = None

    directory: Path

    if not path:
        directory = Path()  # use current Path
    elif path and (not path.suffix or path.is_dir()):
        if not path.is_dir():
            raise ValueError(f"Directory {path} not found. Please create it.")
        directory = path
    else:
        full_path = path
        if not full_path.parent.is_dir():
            raise RuntimeError(
                f"Cannot reach path {path.parent}. Are all directories created?",
            )

    if not full_path:
        _logger.debug("Looking for title")
        prefix = fig.get("layout", {}).get("title", {}).get("text", "fig")
        prefix = re.sub(r"[ \-]", "_", prefix)
        prefix = re.sub(r"[^a-zA-Z0-9_]", "", prefix)
        prefix = prefix or "fig"
        _logger.debug(f"Found: {prefix}")
        name = _next_filename(directory, prefix, ext)
        full_path = directory / name
    return full_path


# call all validators/automatic config fill-in/packaging in expected format
def build_fig_spec(
    fig: Figurish,
    path: Path | str | None,
    opts: LayoutOpts | None,
) -> tuple[Spec, Path]:
    if not opts:
        opts = {}

    if not _is_figurish(fig):
        raise TypeError("Figure supplied doesn't seem to be a valid plotly figure.")

    if hasattr(fig, "to_dict"):
        fig = fig.to_dict()

    if isinstance(path, str):
        path = Path(path)
    elif path and not isinstance(path, Path):
        raise TypeError("Path should be a string or `pathlib.Path` object (or None)")

    if not opts.get("format") and path and path.suffix:
        ext = path.suffix.lstrip(".")
        if _assert_format(ext):  # not strict necessary if but helps typeguard
            opts["format"] = ext

    spec = to_spec(fig, opts)

    full_path = _build_full_path(path, fig, spec["format"])

    return spec, full_path



import asyncio
import re
from unittest.mock import patch

import pytest
from hypothesis import HealthCheck, Phase, given, settings
from hypothesis import strategies as st

from kaleido import Kaleido


@pytest.fixture
async def simple_figure_with_bytes():
    """Create a simple figure with calculated bytes and PNG assertion."""
    import plotly.express as px  # noqa: PLC0415

    fig = px.line(x=[1, 2, 3], y=[1, 2, 3])

    async with Kaleido() as k:
        bytes_data = await k.calc_fig(
            fig,
            opts={"format": "png", "width": 400, "height": 300},
        )

    # Assert it's a PNG by checking the PNG signature
    assert bytes_data[:8] == b"\x89PNG\r\n\x1a\n", "Generated data is not a valid PNG"

    return {
        "fig": fig,
        "bytes": bytes_data,
        "opts": {"format": "png", "width": 400, "height": 300},
    }


async def test_write_fig_from_object_sync_generator(simple_figure_with_bytes, tmp_path):
    """Test write_fig_from_object with sync generator."""

    file_paths = []

    def fig_generator():
        for i in range(2):
            path = tmp_path / f"test_sync_{i}.png"
            file_paths.append(path)
            yield {
                "fig": simple_figure_with_bytes["fig"],
                "path": path,
                "opts": simple_figure_with_bytes["opts"],
            }

    async with Kaleido() as k:
        await k.write_fig_from_object(fig_generator())

    # Assert that each created file matches the fixture bytes
    for path in file_paths:
        assert path.exists(), f"File {path} was not created"
        created_bytes = path.read_bytes()
        assert created_bytes == simple_figure_with_bytes["bytes"], (
            f"File {path} bytes don't match fixture bytes"
        )


async def test_write_fig_from_object_async_generator(
    simple_figure_with_bytes,
    tmp_path,
):
    """Test write_fig_from_object with async generator."""

    file_paths = []

    async def fig_async_generator():
        for i in range(2):
            path = tmp_path / f"test_async_{i}.png"
            file_paths.append(path)
            yield {
                "fig": simple_figure_with_bytes["fig"],
                "path": path,
                "opts": simple_figure_with_bytes["opts"],
            }

    async with Kaleido() as k:
        await k.write_fig_from_object(fig_async_generator())

    # Assert that each created file matches the fixture bytes
    for path in file_paths:
        assert path.exists(), f"File {path} was not created"
        created_bytes = path.read_bytes()
        assert created_bytes == simple_figure_with_bytes["bytes"], (
            f"File {path} bytes don't match fixture bytes"
        )


async def test_write_fig_from_object_iterator(simple_figure_with_bytes, tmp_path):
    """Test write_fig_from_object with iterator."""

    fig_list = []
    file_paths = []
    for i in range(2):
        path = tmp_path / f"test_iter_{i}.png"
        file_paths.append(path)
        fig_list.append(
            {
                "fig": simple_figure_with_bytes["fig"],
                "path": path,
                "opts": simple_figure_with_bytes["opts"],
            },
        )

    async with Kaleido() as k:
        await k.write_fig_from_object(fig_list)

    # Assert that each created file matches the fixture bytes
    for path in file_paths:
        assert path.exists(), f"File {path} was not created"
        created_bytes = path.read_bytes()
        assert created_bytes == simple_figure_with_bytes["bytes"], (
            f"File {path} bytes don't match fixture bytes"
        )


async def test_write_fig_from_object_bare_dictionary(
    simple_figure_with_bytes,
    tmp_path,
):
    """Test write_fig_from_object with bare dictionary list."""

    path1 = tmp_path / "test_dict_1.png"
    path2 = tmp_path / "test_dict_2.png"

    fig_data = [
        {
            "fig": simple_figure_with_bytes["fig"],
            "path": path1,
            "opts": simple_figure_with_bytes["opts"],
        },
        {
            "fig": simple_figure_with_bytes["fig"].to_dict(),
            "path": path2,
            "opts": simple_figure_with_bytes["opts"],
        },
    ]

    async with Kaleido() as k:
        await k.write_fig_from_object(fig_data)

    # Assert that each created file matches the fixture bytes
    for path in [path1, path2]:
        assert path.exists(), f"File {path} was not created"
        created_bytes = path.read_bytes()
        assert created_bytes == simple_figure_with_bytes["bytes"], (
            f"File {path} bytes don't match fixture bytes"
        )


# In the refactor, all figure generation methods are really just wrappers
# for the most flexible, tested above, generate_fig_from_object.
# So we test that one, and then test to make sure its receiving arguments
# properly for the other tests.


# Uncomment these settings after refactor.
# @settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@settings(
    phases=[Phase.generate],
    max_examples=1,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
@given(
    path=st.text(
        min_size=1,
        max_size=50,
        alphabet=st.characters(whitelist_categories=["L", "N"]),
    ),
    width=st.integers(min_value=100, max_value=2000),
    height=st.integers(min_value=100, max_value=2000),
    format_type=st.sampled_from(["png", "svg", "pdf", "html"]),
    topojson=st.one_of(st.none(), st.text(min_size=1, max_size=20)),
)
async def test_write_fig_argument_passthrough(  #  noqa: PLR0913
    simple_figure_with_bytes,
    tmp_path,
    path,
    width,
    height,
    format_type,
    topojson,
):
    """Test that write_fig properly passes arguments to write_fig_from_object."""
    pytest.skip("Remove this failure line and the comment above after the refactor!")
    test_path = tmp_path / f"{path}.{format_type}"
    opts = {"format": format_type, "width": width, "height": height}

    # Mock write_fig_from_object to capture arguments
    with patch.object(Kaleido, "write_fig_from_object") as mock_write_fig_from_object:
        async with Kaleido() as k:
            await k.write_fig(
                simple_figure_with_bytes["fig"],
                path=test_path,
                opts=opts,
                topojson=topojson,
            )

        # Verify write_fig_from_object was called
        mock_write_fig_from_object.assert_called_once()

        # Extract the generator that was passed as first argument
        args, _kwargs = mock_write_fig_from_object.call_args  # not sure.
        assert len(args) == 1, "Expected exactly one argument (the generator)"

        generator = args[0]

        # Convert generator to list to inspect its contents
        generated_args_list = list(generator)
        assert len(generated_args_list) == 1, (
            "Expected generator to yield exactly one item"
        )

        generated_args = generated_args_list[0]

        # Validate that the generated arguments match what we passed to write_fig
        assert "fig" in generated_args, "Generated args should contain 'fig'"
        assert "path" in generated_args, "Generated args should contain 'path'"
        assert "opts" in generated_args, "Generated args should contain 'opts'"
        assert "topojson" in generated_args, "Generated args should contain 'topojson'"

        # Check that the values match
        assert generated_args["fig"] == simple_figure_with_bytes["fig"], (
            "Figure should match"
        )
        assert str(generated_args["path"]) == str(test_path), "Path should match"
        assert generated_args["opts"] == opts, "Options should match"
        assert generated_args["topojson"] == topojson, "Topojson should match"


async def test_kaleido_instantiate_no_hang():
    """Test that instantiating Kaleido doesn't hang."""
    _ = Kaleido()


async def test_kaleido_instantiate_and_close():
    """Test that instantiating and closing Kaleido works."""
    # Maybe there should be a warning or error when closing without opening?
    k = Kaleido()
    await k.close()


async def test_all_methods_context(simple_figure_with_bytes, tmp_path):
    """Test write, write_from_object, and calc with context."""
    fig = simple_figure_with_bytes["fig"]
    opts = simple_figure_with_bytes["opts"]
    expected_bytes = simple_figure_with_bytes["bytes"]

    # Test with context manager
    async with Kaleido() as k:
        # Test calc_fig
        calc_bytes = await k.calc_fig(fig, opts=opts)
        assert calc_bytes == expected_bytes, "calc_fig bytes don't match fixture"

        # Test write_fig
        write_path = tmp_path / "context_write.png"
        await k.write_fig(fig, path=write_path, opts=opts)
        assert write_path.exists(), "write_fig didn't create file"
        write_bytes = write_path.read_bytes()
        assert write_bytes == expected_bytes, "write_fig bytes don't match fixture"

        # Test write_fig_from_object
        obj_path = tmp_path / "context_obj.png"
        await k.write_fig_from_object([{"fig": fig, "path": obj_path, "opts": opts}])
        assert obj_path.exists(), "write_fig_from_object didn't create file"
        obj_bytes = obj_path.read_bytes()
        assert obj_bytes == expected_bytes, (
            "write_fig_from_object bytes don't match fixture"
        )


async def test_all_methods_non_context(simple_figure_with_bytes, tmp_path):
    """Test write, write_from_object, and calc with non-context."""
    fig = simple_figure_with_bytes["fig"]
    opts = simple_figure_with_bytes["opts"]
    expected_bytes = simple_figure_with_bytes["bytes"]

    # Test without context manager
    k = await Kaleido()
    try:
        # Test calc_fig
        calc_bytes = await k.calc_fig(fig, opts=opts)
        assert calc_bytes == expected_bytes, (
            "Non-context calc_fig bytes don't match fixture"
        )

        # Test write_fig

        write_path2 = tmp_path / "non_context_write.png"
        await k.write_fig(fig, path=write_path2, opts=opts)

        assert write_path2.exists(), "Non-context write_fig didn't create file"
        write_bytes2 = write_path2.read_bytes()
        assert write_bytes2 == expected_bytes, (
            "Non-context write_fig bytes don't match fixture"
        )

        obj_path2 = tmp_path / "non_context_obj.png"
        await k.write_fig_from_object([{"fig": fig, "path": obj_path2, "opts": opts}])
        assert obj_path2.exists(), (
            "Non-context write_fig_from_object didn't create file"
        )
        obj_bytes2 = obj_path2.read_bytes()
        assert obj_bytes2 == expected_bytes, (
            "Non-context write_fig_from_object bytes don't match fixture"
        )

    finally:
        await k.close()


@pytest.mark.parametrize("n_tabs", [1, 2, 3])
async def test_tab_count_verification(n_tabs):
    """Test that Kaleido creates the correct number of tabs."""
    async with Kaleido(n=n_tabs) as k:
        # Check the queue size matches expected tabs
        assert k.tabs_ready.qsize() == n_tabs, (
            f"Queue size {k.tabs_ready.qsize()} != {n_tabs}"
        )

        # Use devtools protocol to verify tab count
        # Send getTargets command directly to Kaleido (which is a Browser/Target)
        result = await k.send_command("Target.getTargets")
        # Count targets that are pages (not service workers, etc.)
        page_targets = [
            t for t in result["result"]["targetInfos"] if t.get("type") == "page"
        ]
        assert len(page_targets) >= n_tabs, (
            f"Found {len(page_targets)} page targets, expected at least {n_tabs}"
        )


async def test_unreasonable_timeout(simple_figure_with_bytes):
    """Test that an unreasonably small timeout actually times out."""

    fig = simple_figure_with_bytes["fig"]
    opts = simple_figure_with_bytes["opts"]

    # Use an infinitely small timeout
    async with Kaleido(timeout=0.000001) as k:
        with pytest.raises((asyncio.TimeoutError, TimeoutError)):
            await k.calc_fig(fig, opts=opts)


@pytest.mark.parametrize(
    ("plotlyjs", "mathjax"),
    [
        ("https://cdn.plot.ly/plotly-latest.min.js", None),
        (
            None,
            "https://cdnjs.cloudflare.com/ajax/libs/mathjax/3.2.0/es5/tex-chtml.min.js",
        ),
        (
            "https://cdn.plot.ly/plotly-latest.min.js",
            "https://cdnjs.cloudflare.com/ajax/libs/mathjax/3.2.0/es5/tex-chtml.min.js",
        ),
    ],
)  # THESE STRINGS DON'T ACTUALLY MATTER!
async def test_plotlyjs_mathjax_injection(plotlyjs, mathjax):
    """Test that plotlyjs and mathjax URLs are properly injected."""

    async with Kaleido(plotlyjs=plotlyjs, mathjax=mathjax) as k:
        # Get a tab from the public queue to check the page source
        tab = await k.tabs_ready.get()
        try:
            # Get the page source using devtools protocol
            result = await tab.tab.send_command(
                "Runtime.evaluate",
                {
                    "expression": "document.documentElement.outerHTML",
                },
            )
            source = result["result"]["result"]["value"]

            if plotlyjs:
                # Check if plotlyjs URL is in the source
                plotly_pattern = re.escape(plotlyjs)
                assert re.search(plotly_pattern, source), (
                    f"Plotlyjs URL {plotlyjs} not found in page source"
                )

            if mathjax:
                # Check if mathjax URL is in the source
                mathjax_pattern = re.escape(mathjax)
                assert re.search(mathjax_pattern, source), (
                    f"Mathjax URL {mathjax} not found in page source"
                )

        finally:
            # Put the tab back in the queue
            await k.tabs_ready.put(tab)



"""the kaleido module kaleido.py provides the main classes for the kaleido package."""

from __future__ import annotations

import asyncio
import warnings
from collections.abc import AsyncIterable, Iterable
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import unquote, urlparse

import choreographer as choreo
import logistro
from choreographer.errors import ChromeNotFoundError
from choreographer.utils import TmpDirectory

from ._fig_tools import _is_figurish, build_fig_spec
from ._kaleido_tab import _KaleidoTab
from ._page_generator import PageGenerator
from ._utils import ErrorEntry, warn_incompatible_plotly

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any, Callable, Coroutine

    from . import _fig_tools

_logger = logistro.getLogger(__name__)

try:
    from plotly.utils import PlotlyJSONEncoder  #  noqa: I001
    from choreographer import channels

    channels.register_custom_encoder(PlotlyJSONEncoder)
    _logger.debug("Successfully registered PlotlyJSONEncoder.")
except ImportError as e:
    _logger.debug(f'Couldn\'t import plotly due to "{e!s}" - skipping.')

# Show a warning if the installed Plotly version
# is incompatible with this version of Kaleido
warn_incompatible_plotly()


def _make_printer(name: str) -> Callable[[Any], Coroutine[Any, Any, None]]:
    """Create event printer for generic events. Helper function."""

    async def print_all(response: Any) -> None:
        _logger.debug2(f"{name}:{response}")

    return print_all


class Kaleido(choreo.Browser):
    """
    Kaleido manages a set of image processors.

    It can be used as a context (`async with Kaleido(...)`), but can
    also be used like:

    ```
    k = Kaleido(...)
    k = await Kaleido.open()
    ... # do stuff
    k.close()
    ```
    """

    tabs_ready: asyncio.Queue[_KaleidoTab]
    """A queue of ready tabs."""
    _background_render_tasks: set[asyncio.Task]
    # not really render tasks
    _main_tasks: set[asyncio.Task]

    async def close(self) -> None:
        """Close the browser."""
        await super().close()
        if self._tmp_dir:
            self._tmp_dir.clean()
        _logger.info("Cancelling tasks.")
        for task in self._main_tasks:
            if not task.done():
                task.cancel()
        for task in self._background_render_tasks:
            if not task.done():
                task.cancel()
        _logger.info("Exiting Kaleido/Choreo")

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        """Close the browser."""
        _logger.info("Waiting for all cleanups to finish.")
        await asyncio.gather(*self._background_render_tasks, return_exceptions=True)
        _logger.info("Exiting Kaleido")
        return await super().__aexit__(exc_type, exc_value, exc_tb)

    def __init__(  # noqa: D417, PLR0913 no args/kwargs in description
        self,
        *args: Any,
        page_generator: None | PageGenerator | str | Path = None,
        n: int = 1,
        timeout: int | None = 90,
        width: int | None = None,  # deprecate
        height: int | None = None,  # deprecate
        stepper: bool = False,
        plotlyjs: str | None = None,
        mathjax: str | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize Kaleido, a `choreo.Browser` wrapper adding kaleido functionality.

        It takes all `choreo.Browser` args, plus some extra. The extra
        are listed, see choreographer for more documentation.

        Note: Chrome will throttle background tabs and windows, so non-headless
        multi-process configurations don't work well.

        For argument `page`, if it is a string, it must be passed as a fully-qualified
        URI, like `file://` or `https://`.
        If it is a `Path`, `Path`'s `as_uri()` will be called.
        If it is a string or path, its expected to be an HTML file, one will not
        be generated.

        Args:
            n: the number of separate processes (windows, not seen) to use.
            timeout: limit on any single render (default 90 seconds).
            width: width of window (headless only)
            height: height of window (headless only)
            page: This can be a `kaleido.PageGenerator`, a `pathlib.Path`, or a string.

        """
        self._background_render_tasks = set()
        self._main_tasks = set()
        self.tabs_ready = asyncio.Queue(maxsize=0)
        self._total_tabs = 0
        self._tmp_dir = None

        page = page_generator
        self._timeout = timeout
        self._n = n
        self._height = height  # deprecate
        self._width = width  # deprecate
        self._stepper = stepper
        self._plotlyjs = plotlyjs
        self._mathjax = mathjax
        if not kwargs.get("headless", True) and (self._height or self._width):
            warnings.warn(
                "Height and Width can only be used if headless=True, "
                "ignoring both sizes.",
                stacklevel=1,
            )
            self._height = None
            self._width = None
        _logger.debug(f"Timeout: {self._timeout}")

        try:
            super().__init__(*args, **kwargs)
        except ChromeNotFoundError:
            raise ChromeNotFoundError(
                "Kaleido v1 and later requires Chrome to be installed. "
                "To install Chrome, use the CLI command `kaleido_get_chrome`, "
                "or from Python, use either `kaleido.get_chrome()` "
                "or `kaleido.get_chrome_sync()`.",
            ) from ChromeNotFoundError

        # do this during open because it requires close
        self._saved_page_arg = page

    async def open(self):
        """Build temporary file if we need one."""
        page = self._saved_page_arg
        del self._saved_page_arg

        if isinstance(page, str):
            if page.startswith(r"file://") and Path(unquote(urlparse(page).path)):
                self._index = page
            elif Path(page).is_file():
                self._index = Path(page).as_uri()
            else:
                raise FileNotFoundError(f"{page} does not exist.")
        elif isinstance(page, Path):
            if page.is_file():
                self._index = page.as_uri()
            else:
                raise FileNotFoundError(f"{page!s} does not exist.")
        else:
            self._tmp_dir = TmpDirectory(sneak=self.is_isolated())
            index = self._tmp_dir.path / "index.html"
            self._index = index.as_uri()
            if not page:
                page = PageGenerator(plotly=self._plotlyjs, mathjax=self._mathjax)
            page.generate_index(index)
        await super().open()

    async def _conform_tabs(self, tabs: list[choreo.Tab] | None = None) -> None:
        if not tabs:
            tabs = list(self.tabs.values())
        _logger.info(f"Conforming {len(tabs)} to {self._index}")

        for i, tab in enumerate(tabs):
            n = f"tab-{i!s}"
            _logger.debug2(f"Subscribing * to tab: {tab}.")
            tab.subscribe("*", _make_printer(n + " event"))

        _logger.debug("Navigating all tabs")

        kaleido_tabs = [_KaleidoTab(tab, _stepper=self._stepper) for tab in tabs]

        # A little hard to read because we don't have TaskGroup in this version
        tasks = [asyncio.create_task(tab.navigate(self._index)) for tab in kaleido_tabs]
        _logger.info("Waiting on all navigates")
        await asyncio.gather(*tasks)
        _logger.info("All navigates done, putting them all in queue.")

        for ktab in kaleido_tabs:
            await self.tabs_ready.put(ktab)
        self._total_tabs = len(kaleido_tabs)
        _logger.debug("Tabs fully navigated/enabled/ready")

    async def populate_targets(self) -> None:
        """
        Override the browser populate_targets to ensure the correct page.

        Is called automatically during initialization, and should only be called
        once ever per object.
        """
        await super().populate_targets()
        await self._conform_tabs()
        needed_tabs = self._n - len(self.tabs)
        if needed_tabs < 0:
            raise RuntimeError("Did you set 0 or less tabs?")
        if not needed_tabs:
            return
        tasks = [
            asyncio.create_task(self._create_kaleido_tab()) for _ in range(needed_tabs)
        ]

        await asyncio.gather(*tasks)
        for tab in self.tabs.values():
            _logger.info(f"Tab ready: {tab.target_id}")

    async def _create_kaleido_tab(
        self,
    ) -> None:
        """
        Create a tab with the kaleido script.

        Returns:
            The kaleido-tab created.

        """
        tab = await super().create_tab(
            url="",
            width=self._width,
            height=self._height,
            window=True,
        )
        await self._conform_tabs([tab])

    async def _get_kaleido_tab(self) -> _KaleidoTab:
        """
        Retrieve an available tab from queue.

        Returns:
            A kaleido-tab from the queue.

        """
        _logger.info(f"Getting tab from queue (has {self.tabs_ready.qsize()})")
        if not self._total_tabs:
            raise RuntimeError(
                "Before generating a figure, you must await `k.open()`.",
            )
        tab = await self.tabs_ready.get()
        _logger.info(f"Got {tab.tab.target_id[:4]}")
        return tab

    async def _return_kaleido_tab(self, tab: _KaleidoTab) -> None:
        """
        Refresh tab and put it back into the available queue.

        Args:
            tab: the kaleido tab to return.

        """
        _logger.info(f"Reloading tab {tab.tab.target_id[:4]} before return.")
        await tab.reload()
        _logger.info(
            f"Putting tab {tab.tab.target_id[:4]} back (queue size: "
            f"{self.tabs_ready.qsize()}).",
        )
        await self.tabs_ready.put(tab)
        _logger.debug(f"{tab.tab.target_id[:4]} put back.")

    def _clean_tab_return_task(
        self,
        main_task: asyncio.Task,
        task: asyncio.Task,
    ) -> None:
        _logger.info("Cleaning out background tasks.")
        self._background_render_tasks.remove(task)
        e = task.exception()
        if e:
            _logger.error("Clean tab return task found exception", exc_info=e)
            if not main_task.done():
                main_task.cancel()
            raise e

    def _check_render_task(
        self,
        name: str,
        tab: _KaleidoTab,
        main_task: asyncio.Task,
        error_log: None | list[ErrorEntry],
        task: asyncio.Task,
    ) -> None:
        if task.cancelled():
            _logger.info(f"Something cancelled {name}.")
            if error_log:
                error_log.append(
                    ErrorEntry(name, asyncio.CancelledError, tab.javascript_log),
                )
        elif e := task.exception():
            _logger.error(f"Render Task Error In {name}- ", exc_info=e)
            if isinstance(e, (asyncio.TimeoutError, TimeoutError)) and error_log:
                error_log.append(
                    ErrorEntry(name, e, tab.javascript_log),
                )
            else:
                _logger.error("Cancelling all.")
                if not main_task.done():
                    main_task.cancel()
                raise e
        _logger.info(f"Returning {name} tab after render.")
        t = asyncio.create_task(self._return_kaleido_tab(tab))
        self._background_render_tasks.add(t)
        t.add_done_callback(partial(self._clean_tab_return_task, main_task))

    async def _render_task(
        self,
        tab: _KaleidoTab,
        args: Any,
        error_log: None | list[ErrorEntry] = None,
        profiler: None | list = None,
    ):
        _logger.info(f"Posting a task for {args['full_path'].name}")
        if self._timeout:
            try:
                await asyncio.wait_for(
                    tab._write_fig(  # noqa: SLF001 I don't want it documented, too complex for user
                        **args,
                        error_log=error_log,
                        profiler=profiler,
                    ),
                    self._timeout,  # timeout can be None, no need for branches
                )
            except BaseException as e:
                if error_log:
                    error_log.append(
                        ErrorEntry(
                            args["full_path"].name,
                            e,
                            tab.javascript_log
                            if hasattr(
                                tab,
                                "javascript_log",
                            )
                            else [],
                        ),
                    )
                else:
                    raise

        else:
            await tab._write_fig(  # noqa: SLF001 I don't want it documented, too complex for user
                **args,
                error_log=error_log,
                profiler=profiler,
            )
        _logger.info(f"Posted task ending for {args['full_path'].name}")

    async def calc_fig(
        self,
        fig: _fig_tools.Figurish,
        path: str | Path | None = None,
        opts: None | _fig_tools.LayoutOpts = None,
        *,
        topojson: str | None = None,
    ):
        """
        Calculate the bytes for a figure.

        This function does not support parallelism or multi-image processing like
        `write_fig` does, although its arguments are a subset of those of `write_fig`.
        This function is currently just meant to bridge the old and new API.
        """
        if not _is_figurish(fig) and isinstance(fig, Iterable):
            raise TypeError("Calc fig can not process multiple images at a time.")
        spec, full_path = build_fig_spec(fig, path, opts)
        tab = await self._get_kaleido_tab()
        args = {
            "spec": spec,
            "full_path": full_path,
            "topojson": topojson,
        }
        data = None
        timeout = self._timeout if self._timeout else None
        data = await asyncio.wait_for(
            tab._calc_fig(  # noqa: SLF001 I don't want it documented, too complex for user
                **args,
            ),
            timeout,
        )
        await self._return_kaleido_tab(tab)
        return data[0]

    async def write_fig(  # noqa: PLR0913, PLR0912, C901 (too many args, complexity)
        self,
        fig: _fig_tools.Figurish,
        path: str | Path | None = None,
        opts: _fig_tools.LayoutOpts | None = None,
        *,
        topojson: str | None = None,
        error_log: None | list[ErrorEntry] = None,
        profiler: None | list = None,
    ):
        """
        Call the plotly renderer via javascript on first available tab.

        Args:
            fig: the plotly figure or an iterable of plotly figures
            path: the path to write the images to. if its a directory, we will try to
                generate a name. If the path contains an extension,
                "path/to/my_image.png", that extension will be the format used if not
                overridden in `opts`. If you pass a complete path (filename), for
                multiple figures, you will overwrite every previous figure.
            opts: dictionary describing format, width, height, and scale of image
            topojson: a link ??? TODO
            error_log: a supplied list, will be populated with `ErrorEntry`s
                       which can be converted to strings. Note, this is for
                       collections errors that have to do with plotly. They will
                       not be thrown. Lower level errors (kaleido, choreographer)
                       will still be thrown. If not passed, all errors raise.
            profiler: a supplied dictionary to collect stats about the operation
                      about tabs, runtimes, etc.

        """
        if error_log is not None:
            _logger.info("Using error log.")
        if profiler is not None:
            _logger.info("Using profiler.")

        if _is_figurish(fig) or not isinstance(fig, Iterable):
            fig = [fig]
        else:
            _logger.debug(f"Is iterable {type(fig)}")

        if main_task := asyncio.current_task():
            self._main_tasks.add(main_task)
        tasks = set()

        async def _loop(f):
            spec, full_path = build_fig_spec(f, path, opts)
            tab = await self._get_kaleido_tab()
            if profiler is not None and tab.tab.target_id not in profiler:
                profiler[tab.tab.target_id] = []
            t = asyncio.create_task(
                self._render_task(
                    tab,
                    args={
                        "spec": spec,
                        "full_path": full_path,
                        "topojson": topojson,
                    },
                    error_log=error_log,
                    profiler=profiler,
                ),
            )
            t.add_done_callback(
                partial(
                    self._check_render_task,
                    full_path.name,
                    tab,
                    main_task,
                    error_log,
                ),
            )
            tasks.add(t)

        try:
            if hasattr(fig, "__aiter__"):  # is async iterable
                _logger.debug("Is async for")
                async for f in fig:
                    await _loop(f)
            else:
                _logger.debug("Is sync for")
                for f in fig:
                    await _loop(f)
            _logger.debug("awaiting tasks")
            await asyncio.gather(*tasks, return_exceptions=True)
        except:
            _logger.exception("Cleaning tasks after error.")
            for task in tasks:
                if not task.done():
                    task.cancel()
            raise
        finally:
            if main_task:
                self._main_tasks.remove(main_task)

    async def write_fig_from_object(  # noqa: C901 too complex
        self,
        generator: Iterable | AsyncIterable,
        *,
        error_log: None | list[ErrorEntry] = None,
        profiler: None | list = None,
    ):
        """
        Equal to `write_fig` but allows the user to generate all arguments.

        Generator must yield dictionaries with keys:
        - fig: the plotly figure
        - path: (optional, string or pathlib.Path) the path
        - opts: (optional) dictionary with:
            - format (string)
            - scale (number)
            - height (number)
            - and width (number)
        - topojson: (optional) topojsons are used to customize choropleths

        Generators are good because, if rendering many images, one doesn't need to
        prerender them all. They can be rendered and yielded asynchronously.

        While `write_fig` can also take generators, but only for the figure.
        In this case, the generator will specify all render-related arguments.

        Args:
            generator: an iterable or generator which supplies a dictionary
                       of arguments to pass to tab.write_fig.
            error_log: A supplied list, will be populated with `ErrorEntry`s
                       which can be converted to strings. Note, this is for
                       collections errors that have to do with plotly. They will
                       not be thrown. Lower level errors (kaleido, choreographer)
                       will still be thrown.
            profiler: A supplied dictionary, will be populated with information
                      about tabs, runtimes, etc.

        """
        if error_log is not None:
            _logger.info("Using error log.")
        if profiler is not None:
            _logger.info("Using profiler.")

        if main_task := asyncio.current_task():
            self._main_tasks.add(main_task)
        tasks = set()

        async def _loop(args):
            spec, full_path = build_fig_spec(
                args.pop("fig"),
                args.pop("path", None),
                args.pop("opts", None),
            )
            args["spec"] = spec
            args["full_path"] = full_path
            tab = await self._get_kaleido_tab()
            if profiler is not None and tab.tab.target_id not in profiler:
                profiler[tab.tab.target_id] = []
            t = asyncio.create_task(
                self._render_task(
                    tab,
                    args=args,
                    error_log=error_log,
                    profiler=profiler,
                ),
            )
            t.add_done_callback(
                partial(
                    self._check_render_task,
                    full_path.name,
                    tab,
                    main_task,
                    error_log,
                ),
            )
            tasks.add(t)

        try:
            if hasattr(generator, "__aiter__"):  # is async iterable
                _logger.debug("Is async for")
                async for args in generator:
                    await _loop(args)
            else:
                _logger.debug("Is sync for")
                for args in generator:
                    await _loop(args)
            _logger.debug("awaiting tasks")
            await asyncio.gather(*tasks, return_exceptions=True)
        except:
            _logger.exception("Cleaning tasks after error.")
            for task in tasks:
                if not task.done():
                    task.cancel()
            raise
        finally:
            if main_task:
                self._main_tasks.remove(main_task)



import re
import sys
import tempfile
from html.parser import HTMLParser
from importlib.util import find_spec
from pathlib import Path

import logistro
import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st

from kaleido import PageGenerator
from kaleido._page_generator import DEFAULT_MATHJAX, DEFAULT_PLOTLY

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


# Expected boilerplate HTML (without script tags with src)
EXPECTED_BOILERPLATE = """<!DOCTYPE html>
<html>
    <head>
        <style id="head-style"></style>
        <title>Kaleido-fier</title>
        <script>
          window.PlotlyConfig = {MathJaxConfig: 'local'}
        </script>
        <script type="text/x-mathjax-config">
          MathJax.Hub.Config({ "SVG": { blacker: 0 }})
        </script>

    </head>
    <body style="{margin: 0; padding: 0;}"><img id="kaleido-image" /></body>
</html>"""


class HTMLAnalyzer(HTMLParser):
    """Extract script tags with src attributes and return HTML without them."""

    def __init__(self):
        super().__init__()
        self.scripts = []
        self.encodings = []
        self.boilerplate = []
        self._in_script = False

    def handle_starttag(self, tag, attrs):
        if tag == "script" and "src" in (attr_dict := dict(attrs)):
            self._in_script = True
            self.scripts.append(attr_dict["src"])
            self.encodings.append(attr_dict.get("charset"))
            return
        self.boilerplate.append(self.get_starttag_text())

    def handle_endtag(self, tag):
        if self._in_script and tag == "script":
            self._in_script = False
            return
        self.boilerplate.append(f"</{tag}>")

    def handle_data(self, data):
        if not self._in_script:
            self.boilerplate.append(data)


def normalize_whitespace(html):
    """Normalize whitespace by collapsing multiple newlines and extra spaces."""
    # Collapse multiple newlines to single newlines
    html = re.sub(r"\n\s*\n", "\n", html)
    # Remove extra whitespace between tags
    html = re.sub(r">\s*<", "><", html)
    return html.strip()


# Create boilerplate reference by parsing expected HTML
_reference_analyzer = HTMLAnalyzer()
_reference_analyzer.feed(EXPECTED_BOILERPLATE)
_REFERENCE_BOILERPLATE = normalize_whitespace("".join(_reference_analyzer.boilerplate))


def get_scripts_from_html(generated_html):
    """
    Parse generated HTML, assert boilerplate matches reference, and return script URLs.

    Returns:
        list: script src URLs found in generated HTML
    """
    analyzer = HTMLAnalyzer()
    analyzer.feed(generated_html)

    generated_boilerplate = normalize_whitespace("".join(analyzer.boilerplate))

    # Assert boilerplate matches with diff on failure
    assert generated_boilerplate == _REFERENCE_BOILERPLATE, (
        f"Boilerplate mismatch:\n"
        f"Expected:\n{_REFERENCE_BOILERPLATE}\n\n"
        f"Got:\n{generated_boilerplate}"
    )

    return analyzer.scripts, analyzer.encodings


# Fixtures for user supplied input scenarios
@pytest.fixture
def temp_js_file():
    """Create a temporary JavaScript file that exists."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".js", delete=False) as f:
        f.write("console.log('test');")
        temp_path = Path(f.name)
    yield temp_path
    temp_path.unlink()


@pytest.fixture
def existing_file_path():
    """Return path to current test file (guaranteed to exist)."""
    return Path(__file__)


@pytest.fixture
def nonexistent_file_uri():
    """Return path to file that doesn't exist."""
    return "file:///nonexistent/path/file.js"


@pytest.fixture
def nonexistent_file_path():
    """Return path to file that doesn't exist."""
    return Path("/nonexistent/path/file.js")


def st_valid_path(dir_path: Path):
    file_path = dir_path / "foo.foo"
    file_path.touch()
    assert file_path.resolve().exists()
    _valid_file_string = str(file_path.resolve())

    _h_file_str = st.just(_valid_file_string)
    _h_file_path = st.just(Path(_valid_file_string))
    _h_file_uri = st.just(Path(_valid_file_string).as_uri())

    _h_url = st.tuples(
        st.sampled_from(["s", ""]),
        st.text(
            min_size=1,
            max_size=20,
            alphabet=st.characters(whitelist_categories=("Lu", "Ll")),
        ),
    ).map(lambda x: f"http{x[0]}://example.com/{x[1]}.js")

    _h_uri = st.one_of(_h_url, _h_file_str, _h_file_path, _h_file_uri)

    _h_encoding = st.sampled_from(["utf-8", "utf-16", "ascii", "latin1"])

    return st.one_of(_h_uri, st.tuples(_h_uri, _h_encoding))


# Variable length list strategy for 'others' parameter
def st_others_list(dir_path: Path):
    return st.lists(st_valid_path(dir_path), min_size=0, max_size=3)


# Mathjax strategy (includes None, False, True, and path options)
def st_mathjax(dir_path: Path):
    return st.one_of(
        st.none(),
        st.just(False),  #  noqa: FBT003
        st_valid_path(dir_path),
    )


# Test default combinations
@pytest.mark.order(1)
async def test_defaults_no_plotly_available():
    """
    Test defaults when plotly package is not available.

    When we generate_index(), if we don't have plotly in path, we use a CDN.
    """
    _old_path = sys.path
    try:
        sys.path = []
        _plotly_mo = sys.modules.pop("plotly", None)

        # Test no imports (plotly not available)
        no_imports = PageGenerator().generate_index()
        scripts, _encodings = get_scripts_from_html(no_imports)

        # Should have mathjax, plotly default, and kaleido_scopes
        assert len(scripts) == 3  # noqa: PLR2004
        assert scripts[0] == DEFAULT_MATHJAX
        assert scripts[1] == DEFAULT_PLOTLY
        assert scripts[2].endswith("kaleido_scopes.js")
    finally:
        sys.path = _old_path
        if _plotly_mo:
            sys.modules.update({"plotly": _plotly_mo})


async def test_defaults_with_plotly_available():
    """Test defaults when plotly package is available."""
    all_defaults = PageGenerator().generate_index()
    scripts, _encodings = get_scripts_from_html(all_defaults)

    # Should have mathjax, plotly package data, and kaleido_scopes
    assert len(scripts) == 3  # noqa: PLR2004
    assert scripts[0] == DEFAULT_MATHJAX
    assert scripts[1].endswith("package_data/plotly.min.js")
    assert scripts[2].endswith("kaleido_scopes.js")


async def test_force_cdn():
    """Test force_cdn=True forces use of CDN plotly even when plotly is available."""
    # Verify plotly is available first
    if not find_spec("plotly"):
        pytest.skip("Plotly not available - cannot test force_cdn override")

    forced_cdn = PageGenerator(force_cdn=True).generate_index()
    scripts, _encodings = get_scripts_from_html(forced_cdn)

    assert len(scripts) == 3  # noqa: PLR2004
    assert scripts[0] == DEFAULT_MATHJAX
    assert scripts[1] == DEFAULT_PLOTLY
    assert scripts[2].endswith("kaleido_scopes.js")


# Test boolean mathjax functionality
async def test_mathjax_false():
    """Test that mathjax=False disables mathjax."""
    without_mathjax = PageGenerator(mathjax=False).generate_index()
    scripts, _encodings = get_scripts_from_html(without_mathjax)

    assert len(scripts) == 2  # noqa: PLR2004
    assert scripts[0].endswith("package_data/plotly.min.js")
    assert scripts[1].endswith("kaleido_scopes.js")


# Test user overrides
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(st.data())
async def test_custom_plotly_url(tmp_path, data):
    """Test custom plotly URL override."""
    custom_plotly = data.draw(st_valid_path(tmp_path))
    with_custom = PageGenerator(plotly=custom_plotly).generate_index()
    scripts, encodings = get_scripts_from_html(with_custom)

    assert len(scripts) == 3  # noqa: PLR2004
    assert scripts[0] == DEFAULT_MATHJAX
    if isinstance(custom_plotly, tuple):
        assert scripts[1] == str(custom_plotly[0])
        assert encodings[1] == custom_plotly[1]
    else:
        assert scripts[1] == str(custom_plotly)
    assert scripts[2].endswith("kaleido_scopes.js")


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(st.data())
async def test_custom_mathjax_url(tmp_path, data):
    """Test custom mathjax URL override."""
    custom_mathjax = data.draw(st_valid_path(tmp_path))
    with_custom = PageGenerator(mathjax=custom_mathjax).generate_index()
    scripts, encodings = get_scripts_from_html(with_custom)

    assert len(scripts) == 3  # noqa: PLR2004
    if isinstance(custom_mathjax, tuple):
        assert scripts[0] == str(custom_mathjax[0])
        assert encodings[0] == custom_mathjax[1]
    else:
        assert scripts[0] == str(custom_mathjax)
    assert scripts[1].endswith("package_data/plotly.min.js")
    assert scripts[2].endswith("kaleido_scopes.js")


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(st.data())
async def test_other_scripts(tmp_path, data):
    """Test adding other scripts."""
    other_scripts = data.draw(st_others_list(tmp_path))
    with_others = PageGenerator(others=other_scripts).generate_index()
    scripts, encodings = get_scripts_from_html(with_others)

    # mathjax + plotly + others + kaleido_scopes
    expected_count = 2 + len(other_scripts) + 1
    assert len(scripts) == expected_count
    assert scripts[0] == DEFAULT_MATHJAX
    assert scripts[1].endswith("package_data/plotly.min.js")

    # Check all other scripts in order
    for i, script in enumerate(other_scripts):
        if isinstance(script, tuple):
            assert scripts[2 + i] == str(script[0])
            assert encodings[2 + i] == script[1]
        else:
            assert scripts[2 + i] == str(script)

    assert scripts[-1].endswith("kaleido_scopes.js")


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(st.data())
async def test_combined_overrides(tmp_path, data):
    """Test combination of multiple overrides."""
    custom_plotly = data.draw(st_valid_path(tmp_path))
    custom_mathjax = data.draw(st_mathjax(tmp_path))
    other_scripts = data.draw(st_others_list(tmp_path))

    combined = PageGenerator(
        plotly=custom_plotly,
        mathjax=custom_mathjax,
        others=other_scripts,
    ).generate_index()
    scripts, encodings = get_scripts_from_html(combined)

    # Calculate expected count
    expected_count = 0
    script_index = 0

    # Mathjax adds one (unless False)
    if custom_mathjax is not False:
        expected_count += 1
        if custom_mathjax is None:
            expected_mathjax = DEFAULT_MATHJAX
        elif isinstance(custom_mathjax, tuple):
            expected_mathjax = str(custom_mathjax[0])
            assert encodings[script_index] == custom_mathjax[1]
        else:
            expected_mathjax = str(custom_mathjax)
        assert scripts[script_index] == expected_mathjax
        script_index += 1

    # Plotly always adds one
    expected_count += 1
    if isinstance(custom_plotly, tuple):
        assert scripts[script_index] == str(custom_plotly[0])
        assert encodings[script_index] == custom_plotly[1]
    else:
        assert scripts[script_index] == str(custom_plotly)
    script_index += 1

    # Others adds however many are in the list
    expected_count += len(other_scripts)
    for script in other_scripts:
        if isinstance(script, tuple):
            assert scripts[script_index] == str(script[0])
            assert encodings[script_index] == script[1]
        else:
            assert scripts[script_index] == str(script)
        script_index += 1

    # Kaleido scopes always adds one
    expected_count += 1
    assert scripts[script_index].endswith("kaleido_scopes.js")

    assert len(scripts) == expected_count


# Test file path validation
async def test_existing_file_path(temp_js_file):
    """Test that existing file paths work with and without file:/// protocol."""
    # Test with regular path
    generator = PageGenerator(plotly=str(temp_js_file))
    html = generator.generate_index()
    scripts, _encodings = get_scripts_from_html(html)
    assert len(scripts) == 3  # noqa: PLR2004
    assert scripts[0] == DEFAULT_MATHJAX
    assert scripts[1] == str(temp_js_file)
    assert scripts[2].endswith("kaleido_scopes.js")

    # Test with file:/// protocol
    generator_uri = PageGenerator(plotly=temp_js_file.as_uri())
    html_uri = generator_uri.generate_index()
    scripts_uri, _encodings_uri = get_scripts_from_html(html_uri)
    assert len(scripts_uri) == 3  # noqa: PLR2004
    assert scripts_uri[0] == DEFAULT_MATHJAX
    assert scripts_uri[1] == temp_js_file.as_uri()
    assert scripts_uri[2].endswith("kaleido_scopes.js")


async def test_nonexistent_file_path_raises_error(
    nonexistent_file_path,
    nonexistent_file_uri,
):
    """Test that nonexistent file paths raise FileNotFoundError."""
    # Test with regular path
    with pytest.raises(FileNotFoundError):
        PageGenerator(plotly=str(nonexistent_file_path))

    with pytest.raises(FileNotFoundError):
        PageGenerator(plotly=Path(nonexistent_file_path))

    # Test with file:/// protocol
    with pytest.raises(FileNotFoundError):
        PageGenerator(plotly=nonexistent_file_uri)


async def test_mathjax_nonexistent_file_raises_error(
    nonexistent_file_path,
    nonexistent_file_uri,
):
    """Test that nonexistent mathjax file raises FileNotFoundError."""
    # Test with regular path
    with pytest.raises(FileNotFoundError):
        PageGenerator(mathjax=str(nonexistent_file_path))

    with pytest.raises(FileNotFoundError):
        PageGenerator(mathjax=nonexistent_file_path)

    # Test with file:/// protocol
    with pytest.raises(FileNotFoundError):
        PageGenerator(mathjax=nonexistent_file_uri)


async def test_others_nonexistent_file_raises_error(
    nonexistent_file_path,
    nonexistent_file_uri,
):
    """Test that nonexistent file in others list raises FileNotFoundError."""
    # Test with regular path
    with pytest.raises(FileNotFoundError):
        PageGenerator(others=[str(nonexistent_file_path)])

    with pytest.raises(FileNotFoundError):
        PageGenerator(others=[nonexistent_file_path])

    # Test with file:/// protocol
    with pytest.raises(FileNotFoundError):
        PageGenerator(others=[nonexistent_file_uri])


# Test HTTP URLs (should not raise FileNotFoundError)
async def test_http_urls_skip_file_validation():
    """Test that HTTP URLs skip file existence validation."""
    # These should not raise FileNotFoundError even if URLs don't exist
    generator = PageGenerator(
        plotly="https://nonexistent.example.com/plotly.js",
        mathjax="https://nonexistent.example.com/mathjax.js",
        others=["https://nonexistent.example.com/other.js"],
    )
    html = generator.generate_index()
    scripts, _encodings = get_scripts_from_html(html)

    assert len(scripts) == 4  # noqa: PLR2004
    assert scripts[0] == "https://nonexistent.example.com/mathjax.js"
    assert scripts[1] == "https://nonexistent.example.com/plotly.js"
    assert scripts[2] == "https://nonexistent.example.com/other.js"
    assert scripts[3].endswith("kaleido_scopes.js")



import warnings

import logistro
import pytest

import kaleido

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


async def test_calc_fig():
    # ruff: noqa: PLC0415
    import plotly.express as px

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        fig = px.line(x=[1, 2, 3, 4], y=[1, 2, 3, 4])

    # Pixel by pixel testing
    img_awaited = await kaleido.calc_fig(fig)
    assert isinstance(img_awaited, bytes)

    # Make sure passing `mathjax` argument via `kopts` works properly
    img_sync = kaleido.calc_fig_sync(
        fig,
        kopts={"mathjax": "https://cdn.jsdelivr.net/npm/mathjax@3.2.1/es5/tex-svg.js"},
    )
    assert isinstance(img_sync, bytes)

    img_from_dict = kaleido.calc_fig_sync(fig.to_dict())
    assert isinstance(img_from_dict, bytes)

    assert img_awaited == img_sync == img_from_dict

    with pytest.raises(TypeError):
        # can't accept iterables
        _ = kaleido.calc_fig_sync([fig, fig])



import asyncio
import traceback
import warnings
from functools import partial
from importlib.metadata import PackageNotFoundError, version

import logistro
from packaging.version import Version

_logger = logistro.getLogger(__name__)


async def to_thread(func, *args, **kwargs):
    _loop = asyncio.get_running_loop()
    fn = partial(func, *args, **kwargs)
    await _loop.run_in_executor(None, fn)


def warn_incompatible_plotly():
    """
    Check if installed Plotly version (if any) is compatible with this Kaleido version.

    If not, display a warning.
    """
    try:
        min_compatible_plotly_version = Version("6.1.1")
        installed_plotly_version = Version(version("plotly"))
        installed_kaleido_version = Version(version("kaleido"))
        if installed_plotly_version < min_compatible_plotly_version:
            warnings.warn(
                "\n\n"
                f"Warning: You have Plotly version {installed_plotly_version}, "
                "which is not compatible with this version of "
                f"Kaleido ({installed_kaleido_version}).\n\n"
                "This means that static image generation (e.g. `fig.write_image()`) "
                "will not work.\n\n"
                f"Please upgrade Plotly to version {min_compatible_plotly_version} "
                "or greater, or downgrade Kaleido to version 0.2.1."
                "\n",
                UserWarning,
                stacklevel=3,
            )
    except PackageNotFoundError:
        # If Plotly is not installed, there's nothing to worry about
        pass
    # ruff: noqa: BLE001
    except Exception as e:
        # If another error occurs, log it but do not raise
        # Since this compatibility check is just a convenience,
        # we don't want to block the whole library if there's an issue
        _logger.info("Error while checking Plotly version.", exc_info=e)


class ErrorEntry:
    """A simple object to record errors and context."""

    def __init__(self, name, error, javascript_log):
        """
        Construct an error entry.

        Args:
            name: the name of the image with the error
            error: the error object (from class BaseException)
            javascript_log: an array of entries from the javascript console

        """
        self.name = name
        self.error = error
        self.javascript_log = javascript_log

    def __str__(self):
        """Display the error object in a concise way."""
        ret = f"{self.name}:\n"
        e = self.error
        ret += " ".join(traceback.format_exception(type(e), e, e.__traceback__))
        ret += " javascript Log:\n"
        ret += "\n ".join(self.javascript_log)
        return ret



from __future__ import annotations

import base64
import json
import time
from typing import TYPE_CHECKING

import logistro
from choreographer.errors import DevtoolsProtocolError

from ._utils import ErrorEntry, to_thread

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    import choreographer as choreo

_logger = logistro.getLogger(__name__)

_TEXT_FORMATS = ("svg", "json")  # eps


class JavascriptError(RuntimeError):  # TODO(A): process better # noqa: TD003, FIX002
    """Used to report errors from javascript."""


### Error definitions ###
class KaleidoError(Exception):
    """An error to interpret errors from Kaleido's JS side."""

    def __init__(self, code, message):
        """
        Construct an error object.

        Args:
            code: the number code of the error.
            message: the message of the error.

        """
        super().__init__(message)
        self._code = code
        self._message = message

    def __str__(self):
        """Display the KaleidoError nicely."""
        return f"Error {self._code}: {self._message}"


def _check_error(result):
    e = _check_error_ret(result)
    if e:
        raise e


def _check_error_ret(result):  # Utility
    """Check browser response for errors. Helper function."""
    if "error" in result:
        return DevtoolsProtocolError(result)
    if result.get("result", {}).get("result", {}).get("subtype", None) == "error":
        return JavascriptError(str(result.get("result")))
    return None


def _make_console_logger(name, log):
    """Create printer specifically for console events. Helper function."""

    async def console_printer(event):
        _logger.debug2(f"{name}:{event}")  # TODO(A): parse # noqa: TD003, FIX002
        log.append(str(event))

    return console_printer


class _KaleidoTab:
    """
    A Kaleido tab is a wrapped choreographer tab providing the functions we need.

    The choreographer tab can be access through the `self.tab` attribute.
    """

    tab: choreo.Tab
    """The underlying choreographer tab."""

    javascript_log: list[Any]
    """A list of console outputs from the tab."""

    def __init__(self, tab, *, _stepper=False):
        """
        Create a new _KaleidoTab.

        Args:
            tab: the choreographer tab to wrap.

        """
        self.tab = tab
        self.javascript_log = []
        self._stepper = _stepper

    def _regenerate_javascript_console(self):
        tab = self.tab
        self.javascript_log = []
        _logger.debug2("Subscribing to all console prints for tab {tab}.")
        tab.unsubscribe("Runtime.consoleAPICalled")
        tab.subscribe(
            "Runtime.consoleAPICalled",
            _make_console_logger("tab js console", self.javascript_log),
        )

    async def navigate(self, url: str | Path = ""):
        """
        Navigate to the kaleidofier script. This is effectively the real initialization.

        Args:
            url: Override the location of the kaleidofier script if necessary.

        """
        tab = self.tab
        javascript_ready = tab.subscribe_once("Runtime.executionContextCreated")
        while javascript_ready.done():
            _logger.debug2("Clearing an old Runtime.executionContextCreated")
            javascript_ready = tab.subscribe_once("Runtime.executionContextCreated")
        page_ready = tab.subscribe_once("Page.loadEventFired")
        while page_ready.done():
            _logger.debug2("Clearing a old Page.loadEventFired")
            page_ready = tab.subscribe_once("Page.loadEventFired")

        _logger.debug2(f"Calling Page.navigate on {tab}")
        _check_error(await tab.send_command("Page.navigate", params={"url": url}))
        # Must enable after navigating.
        _logger.debug2(f"Calling Page.enable on {tab}")
        _check_error(await tab.send_command("Page.enable"))
        _logger.debug2(f"Calling Runtime.enable on {tab}")
        _check_error(await tab.send_command("Runtime.enable"))

        await javascript_ready
        self._current_js_id = (
            javascript_ready.result()
            .get("params", {})
            .get("context", {})
            .get("id", None)
        )
        if not self._current_js_id:
            raise RuntimeError(
                "Refresh sequence didn't work for reload_tab_with_javascript."
                "Result {javascript_ready.result()}.",
            )
        await page_ready
        self._regenerate_javascript_console()

    async def reload(self):
        """Reload the tab, and set the javascript runtime id."""
        tab = self.tab
        _logger.debug(f"Reloading tab {tab} with javascript.")
        javascript_ready = tab.subscribe_once("Runtime.executionContextCreated")
        while javascript_ready.done():
            _logger.debug2("Clearing an old Runtime.executionContextCreated")
            javascript_ready = tab.subscribe_once("Runtime.executionContextCreated")
        is_loaded = tab.subscribe_once("Page.loadEventFired")
        while is_loaded.done():
            _logger.debug2("Clearing an old Page.loadEventFired")
            is_loaded = tab.subscribe_once("Page.loadEventFired")
        _logger.debug2(f"Calling Page.reload on {tab}")
        _check_error(await tab.send_command("Page.reload"))
        await javascript_ready
        self._current_js_id = (
            javascript_ready.result()
            .get("params", {})
            .get("context", {})
            .get("id", None)
        )
        if not self._current_js_id:
            raise RuntimeError(
                "Refresh sequence didn't work for reload_tab_with_javascript."
                "Result {javascript_ready.result()}.",
            )
        await is_loaded
        self._regenerate_javascript_console()

    async def console_print(self, message: str) -> None:
        """
        Print something to the javascript console.

        Args:
            message: The thing to print.

        """
        jsfn = r"function()" r"{" f"console.log('{message}')" r"}"
        params = {
            "functionDeclaration": jsfn,
            "returnByValue": False,
            "userGesture": True,
            "awaitPromise": True,
            "executionContextId": self._current_js_id,
        }

        # send request to run script in chromium
        _logger.debug("Calling js function")
        result = await self.tab.send_command("Runtime.callFunctionOn", params=params)
        _logger.debug(f"Sent javascript got result: {result}")
        _check_error(result)

    def _finish_profile(self, profile, state, error=None):
        _logger.debug("Finishing profile")
        profile["duration"] = float(f"{time.perf_counter() - profile['start']:.6f}")
        del profile["start"]
        profile["state"] = state
        if self.javascript_log:
            profile["js_console"] = self.javascript_log
        if error:
            profile["error"] = error

    async def _write_fig(
        self,
        spec,
        full_path,
        *,
        topojson=None,
        error_log=None,
        profiler=None,
    ):
        """Calculate and write figure to file. Wraps _calc_fig, and writes a file."""
        img, profile = await self._calc_fig(
            spec,
            full_path,
            topojson=topojson,
            error_log=error_log,
            profiler=profiler,
        )

        def write_image(binary):
            with full_path.open("wb") as file:
                file.write(binary)

        _logger.info(f"Starting write of {full_path.name}")
        await to_thread(write_image, img)
        _logger.info(f"Wrote {full_path.name}")

        if profile is not None:
            profile["megabytes"] = full_path.stat().st_size / 1000000
            profile["state"] = "WROTE"

    async def _calc_fig(  # noqa: C901, PLR0912, complexity, branches
        self,
        spec,
        full_path,
        *,
        topojson=None,
        error_log=None,
        profiler=None,
    ):
        """
        Call the plotly renderer via javascript.

        Args:
            spec: the processed plotly figure
            full_path: the path to write the image too. if its a directory, we will try
                to generate a name. If the path contains an extension,
                "path/to/my_image.png", that extension will be the format used if not
                overridden in `opts`.
            opts: dictionary describing format, width, height, and scale of image
            topojson: topojsons are used to customize choropleths
            error_log: A supplied list, will be populated with `ErrorEntry`s
                       which can be converted to strings. Note, this is for
                       collections errors that have to do with plotly. They will
                       not be thrown. Lower level errors (kaleido, choreographer)
                       will still be thrown. If not passed, all errors raise.
            profiler: a supplied dictionary to collect stats about the operation

        """
        tab = self.tab
        execution_context_id = self._current_js_id
        if profiler is not None:
            profile = {
                "name": full_path.name,
                "start": time.perf_counter(),
                "state": "INIT",
            }
        else:
            profile = None

        _logger.debug(f"In tab {tab.target_id[:4]} calc_fig for {full_path.name}.")

        _logger.info(f"Processing {full_path.name}")
        # js script
        kaleido_jsfn = (
            r"function(spec, ...args)"
            r"{"
            r"return kaleido_scopes.plotly(spec, ...args).then(JSON.stringify);"
            r"}"
        )

        # params
        arguments = [{"value": spec}]
        arguments.append({"value": topojson if topojson else None})
        arguments.append({"value": self._stepper})
        params = {
            "functionDeclaration": kaleido_jsfn,
            "arguments": arguments,
            "returnByValue": False,
            "userGesture": True,
            "awaitPromise": True,
            "executionContextId": execution_context_id,
        }

        _logger.info(f"Sending big command for {full_path.name}.")
        if profile:
            profile["state"] = "SENDING"
        result = await tab.send_command("Runtime.callFunctionOn", params=params)
        if profile:
            profile["state"] = "SENT"
        _logger.info(f"Sent big command for {full_path.name}.")
        e = _check_error_ret(result)
        if e:
            if profiler is not None:
                self._finish_profile(profile, "ERROR", e)
                profiler[tab.target_id].append(profile)
            if error_log is not None:
                error_log.append(ErrorEntry(full_path.name, e, self.javascript_log))
                _logger.error(f"Failed {full_path.name}", exc_info=e)
            else:
                _logger.error(f"Raising error on {full_path.name}")
                raise e
        _logger.debug2(f"Result of function call: {result}")
        if self._stepper:
            print(f"Image {full_path.name} was sent to browser")  # noqa: T201
            input("Press Enter to continue...")
        if e:
            return None, None

        img = await self._img_from_response(result)
        if isinstance(img, BaseException):
            if profiler is not None:
                self._finish_profile(profile, "ERROR", img)
                profiler[tab.target_id].append(profile)
            if error_log is not None:
                error_log.append(
                    ErrorEntry(full_path.name, img, self.javascript_log),
                )
                _logger.info(f"Failed {full_path.name}")
                return None, None
            else:
                raise img
        if profile:
            self._finish_profile(profile, "CALCULATED", None)
            profiler[tab.target_id].append(profile)
        return img, profile

    async def _img_from_response(self, response):
        js_response = json.loads(response.get("result").get("result").get("value"))

        if js_response["code"] != 0:
            return KaleidoError(js_response["code"], js_response["message"])

        response_format = js_response.get("format")
        img = js_response.get("result")
        if response_format == "pdf":
            pdf_params = {
                "printBackground": True,
                "marginTop": 0.1,
                "marginBottom": 0.1,
                "marginLeft": 0.1,
                "marginRight": 0.1,
                "preferCSSPageSize": True,
                "pageRanges": "1",
            }
            pdf_response = await self.tab.send_command(
                "Page.printToPDF",
                params=pdf_params,
            )
            e = _check_error_ret(pdf_response)
            if e:
                return e
            img = pdf_response.get("result").get("data")
        # Base64 decode binary types
        if response_format not in _TEXT_FORMATS:
            img = base64.b64decode(img)
        else:
            img = str.encode(img)
        return img



from __future__ import annotations

import asyncio
import atexit
import warnings
from functools import partial
from queue import Queue
from threading import Thread
from typing import TYPE_CHECKING, NamedTuple

from .kaleido import Kaleido

if TYPE_CHECKING:
    from typing import Any


class Task(NamedTuple):
    fn: str
    args: Any
    kwargs: Any


class _BadFunctionName(BaseException):
    """For use when programmed poorly."""


class GlobalKaleidoServer:
    _instance = None

    async def _server(self, *args, **kwargs):
        async with Kaleido(*args, **kwargs) as k:  # multiple processor? Enable GPU?
            while True:
                task = self._task_queue.get()  # thread dies if main thread dies
                if task is None:
                    self._task_queue.task_done()
                    return
                if not hasattr(k, task.fn):
                    raise _BadFunctionName(f"Kaleido has no attribute {task.fn}")
                try:
                    self._return_queue.put(
                        await getattr(k, task.fn)(*task.args, **task.kwargs),
                    )
                except Exception as e:  # noqa: BLE001
                    self._return_queue.put(e)

                self._task_queue.task_done()

    def __new__(cls):
        # Create the singleton on first instantiation
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False  # noqa: SLF001
        return cls._instance

    def is_running(self):
        return self._initialized

    def open(self, *args: Any, silence_warnings=False, **kwargs: Any) -> None:
        """Initialize the singleton with three values."""
        if self.is_running():
            if not silence_warnings:
                warnings.warn(
                    "Server already open.",
                    RuntimeWarning,
                    stacklevel=2,
                )
            return
        coroutine = self._server(*args, **kwargs)
        self._thread: Thread = Thread(
            target=asyncio.run,
            args=(coroutine,),
            daemon=True,
        )
        self._task_queue: Queue[Task | None] = Queue()
        self._return_queue: Queue[Any] = Queue()
        self._thread.start()
        self._initialized = True
        close = partial(self.close, silence_warnings=True)
        atexit.register(close)

    def close(self, *, silence_warnings=False):
        """Reset the singleton back to an uninitialized state."""
        if not self.is_running():
            if not silence_warnings:
                warnings.warn(
                    "Server already closed.",
                    RuntimeWarning,
                    stacklevel=2,
                )
            return
        self._task_queue.put(None)
        self._thread.join()
        del self._thread
        del self._task_queue
        del self._return_queue
        self._initialized = False

    def call_function(self, cmd: str, *args, **kwargs):
        if not self.is_running():
            raise RuntimeError("Can't call function on stopped server.")
        if kwargs.pop("kopts", None):
            warnings.warn(
                "The kopts argument is ignored if using a server.",
                UserWarning,
                stacklevel=3,
            )
        self._task_queue.put(Task(cmd, args, kwargs))
        self._task_queue.join()
        res = self._return_queue.get()
        if isinstance(res, BaseException):
            raise res
        else:
            return res


def oneshot_async_run(func, args: tuple[Any, ...], kwargs: dict):
    q: Queue[Any] = Queue(maxsize=1)

    def run(func, q, *args, **kwargs):
        # func is a closure
        try:
            q.put(asyncio.run(func(*args, **kwargs)))
        except BaseException as e:  # noqa: BLE001
            q.put(e)

    t = Thread(target=run, args=(func, q, *args), kwargs=kwargs)
    t.start()
    t.join()
    res = q.get()
    if isinstance(res, BaseException):
        raise res
    else:
        return res
